/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <aul_cmd.h>

#include "component_manager.h"
#include "request_handler/component_info_get.h"
#include "request_handler/component_info_foreach.h"
#include "private.h"

namespace component_manager {
using namespace component_manager::request_handler;

int ComponentManager::Register() {
  int ret = RegisterCommands();
  if (ret< 0)
    return ret;

  ret = RegisterCynaraCheckers();
  if (ret < 0)
    return ret;

  return 0;
}

int ComponentManager::Dispatch(std::unique_ptr<Request> req) {
  return dispatcher_->Dispatch(std::move(req));
}

void ComponentManager::Init() {
  dispatcher_ = std::unique_ptr<Dispatcher>(new (std::nothrow) Dispatcher());
  dispatcher_->Register(std::shared_ptr<RequestHandler>(
        new (std::nothrow) ComponentInfoGet()));
  dispatcher_->Register(std::shared_ptr<RequestHandler>(
        new (std::nothrow) ComponentInfoForeach()));

  initialized_ = true;
}

int ComponentManager::RegisterCommands() {
  static amd_request_cmd_dispatch dispatch_table[] = {
    {
      .cmd = COMP_INFO_GET,
      .callback = OnDispatch
    },
    {
      .cmd = COMP_INFO_FOREACH,
      .callback = OnDispatch
    },
  };

  int ret = amd_request_register_cmds(dispatch_table,
      ARRAY_SIZE(dispatch_table));
  if (ret < 0) {
    _E("Failed to register AMD request cmds");
    return ret;
  }

  return 0;
}

int ComponentManager::RegisterCynaraCheckers() {
  void* packagemanager_info = const_cast<char*>(PRIVILEGE_PACKAGEMANAGER_INFO);
  static amd_cynara_checker cynara_checkers[] = {
    {
      .cmd = COMP_INFO_GET,
      .checker = amd_cynara_simple_checker,
      .data = packagemanager_info
    },
    {
      .cmd = COMP_INFO_FOREACH,
      .checker = amd_cynara_simple_checker,
      .data = packagemanager_info
    },
  };

  int ret = amd_cynara_register_checkers(cynara_checkers,
      ARRAY_SIZE(cynara_checkers));
  if (ret < 0) {
    _E("Failed to register cynara checkers");
    return ret;
  }

  return 0;
}

int ComponentManager::OnDispatch(amd_request_h request) {
  auto req = std::unique_ptr<Request>(new (std::nothrow) Request(request));
  auto& inst = ComponentManager::GetInst();
  if (inst.Dispatch(std::move(req)) < 0) {
    _E("Failed to dispatch request");
    return -1;
  }

  return 0;
}

}  // namespace component_manager
