/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <aul.h>
#include <aul_comp_info_internal.h>
#include <aul_cmd.h>
#include <errno.h>

#ifdef _GNU_SOURCE
#undef _GNU_SOURCE
#endif
#include <aul_sock.h>
#include <amd.h>

#include "component_info.h"
#include "request_handler/component_info_foreach.h"
#include "private.h"

namespace component_manager {
namespace request_handler {

class ComponentInfoForeachUserData {
 public:
  ComponentInfoForeachUserData(uid_t uid, void* user_data)
    : uid_(uid), user_data_(user_data) { }

  uid_t uid_;
  void* user_data_;
};

ComponentInfoForeach::ComponentInfoForeach()
  : RequestHandler(COMP_INFO_FOREACH, "COMP_INFO_FOREACH") {
}

int ComponentInfoForeach::Handle(std::unique_ptr<Request>& req) {
  RequestHandler::Handle(req);

  std::vector<std::unique_ptr<ComponentInfo>> infos;
  ComponentInfoForeachUserData user_data(req->GetTargetUid(), &infos);
  int ret = aul_compinfo_usr_foreach_compinfo(req->GetTargetUid(),
      [](aul_compinfo_h handle, void* user_data) -> bool {
        aul_compinfo_h clone = nullptr;
        aul_compinfo_clone(handle, &clone);
        if (clone == nullptr) {
          _E("Failed to clone compinfo");
          return false;
        }

        auto* data = static_cast<ComponentInfoForeachUserData*>(user_data);
        auto* infos = static_cast<std::vector<std::unique_ptr<ComponentInfo>>*>(
            data->user_data_);
        infos->emplace_back(
            new (std::nothrow) ComponentInfo(clone, data->uid_));
        return true;
      }, &user_data);
  if (ret < 0) {
    _E("Failed to retrieve component info. error(%d)");
    amd_socket_send_result(req->GetFd(), ret, false);
    return ret;
  }

  amd_socket_send_result(req->GetFd(), infos.size(), false);
  for (auto& info : infos) {
    auto b = info->ToBundle();
    bundle* b_raw = b.GetHandle();
    ret = aul_sock_send_bundle_with_fd(req->GetFd(), APP_GET_INFO_OK, b_raw,
        AUL_SOCK_ASYNC);
    if (ret < 0) {
      _E("Failed to send bundle data. error(%d)", ret);
      return ret;
    }
  }

  _I("[%s][%d]", GetName().c_str(), GetCmd());
  return 0;
}

}  // namespace request_handler
}  // namespace component_manager
