/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <glib.h>
#include <gio/gio.h>
#include <aul.h>
#include <amd.h>

#include "amd_cooldown.h"

#define DBUS_SERVICE_DBUS "org.freedesktop.DBus"
#define DBUS_PATH_DBUS "/org/freedesktop/DBus"
#define DBUS_INTERFACE_DBUS "org.freedesktop.DBus"

#define AMD_DBUS_INTERFACE "org.tizen.application.manager"
#define AMD_DBUS_OBJECT_PATH "/Org/Tizen/Application/Manager"

#define REGULAR_UID_MIN 5000

#define COOLDOWN_STATUS_RELEASE "release"
#define COOLDOWN_STATUS_LIMIT "limit"

enum cooldown_status_val {
	COOLDOWN_RELEASE,
	COOLDOWN_WARNING,
	COOLDOWN_LIMIT,
};

static int cooldown_status;
static GDBusConnection *__conn;
static guint __registration_id;

static void __cooldown_limitaction(amd_app_status_h app_status, void *data)
{
	amd_appinfo_h ai;
	const char *taskmanage;
	const char *cooldown;
	const char *appid;
	int app_type;
	uid_t uid;

	if (app_status == NULL)
		return;

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type == AMD_AT_WIDGET_APP || app_type == AMD_AT_WATCH_APP)
		return;

	uid = amd_app_status_get_uid(app_status);
	appid = amd_app_status_get_appid(app_status);
	ai = amd_appinfo_find(uid, appid);
	if (ai == NULL)
		return;

	cooldown = amd_appinfo_get_value(ai, AMD_AIT_COOLDOWN);
	if (cooldown && strcmp(cooldown, "true") != 0) {
		if (app_type == AMD_AT_UI_APP) {
			taskmanage = amd_appinfo_get_value(ai,
					AMD_AIT_TASKMANAGE);
			if (taskmanage && strcmp(taskmanage, "true") != 0)
				return;
		}

		amd_app_status_terminate_apps(appid, uid);
	}
}

static void __cooldown_release(void)
{
	amd_uid_state state;
	uid_t *uids;
	int r;
	int i;

	r = amd_login_monitor_get_uids(&uids);
	if (r <= 0)
		return;

	for (i = 0; i < r; ++i) {
		state = amd_login_monitor_get_uid_state(uids[i]);
		if (state == AMD_UID_STATE_ACTIVE)
			amd_launch_start_onboot_apps(uids[i]);
	}
	free(uids);
}

static int __check_cooldown_mode(amd_appinfo_h ai)
{
	const char *taskmanage;
	const char *cooldown;
	const char *comptype;

	if (cooldown_status != COOLDOWN_LIMIT)
		return 0;

	cooldown = amd_appinfo_get_value(ai, AMD_AIT_COOLDOWN);
	if (cooldown && strcmp(cooldown, "true") == 0)
		return 0;

	comptype = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);
	if (comptype == NULL) {
		_E("Failed to get comptype");
		return -1;
	}

	if (strcmp(comptype, APP_TYPE_WIDGET) == 0 ||
			strcmp(comptype, APP_TYPE_WATCH) == 0)
		return 0;

	if (strcmp(comptype, APP_TYPE_UI) == 0) {
		taskmanage = amd_appinfo_get_value(ai, AMD_AIT_TASKMANAGE);
		if (taskmanage && strcmp(taskmanage, "false") == 0)
			return 0;
	}

	_W("Cannot launch this application in COOLDOWN mode");
	return -1;
}

static int __on_check_mode(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	amd_appinfo_h info = arg3;

	return __check_cooldown_mode(info);
}

static int __on_check_status(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	if (cooldown_status == COOLDOWN_LIMIT)
		return -1;

	return 0;
}

static uid_t __get_caller_uid(GDBusConnection *connection, const char *name)
{
	GError *err = NULL;
	GVariant *result;
	uid_t uid = (uid_t)-1;

	result = g_dbus_connection_call_sync(connection,
			DBUS_SERVICE_DBUS,
			DBUS_PATH_DBUS,
			DBUS_INTERFACE_DBUS,
			"GetConnectionUnixUser",
			g_variant_new("(s)", name),
			NULL,
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&err);
	if (result == NULL) {
		_E("g_dbus_connection_call_sync() is failed. error(%s)",
				err->message);
		g_error_free(err);
		return uid;
	}

	g_variant_get(result, "(u)", &uid);
	g_variant_unref(result);

	return uid;
}

static void __handle_method_call(GDBusConnection *connection,
		const gchar *sender,
		const gchar *object_path,
		const gchar *interface_name,
		const gchar *method_name,
		GVariant *parameters,
		GDBusMethodInvocation *invocation,
		gpointer user_data)
{
	int result = 0;
	char *status = NULL;
	uid_t uid;

	uid = __get_caller_uid(connection,
			g_dbus_method_invocation_get_sender(invocation));
	if (uid == (uid_t)-1) {
		result = -1;
		goto end;
	}

	if (uid >= REGULAR_UID_MIN) {
		_E("Reject request. uid(%u)", uid);
		result = -1;
		goto end;
	}

	if (g_strcmp0(method_name, "Cooldown") == 0) {
		g_variant_get(parameters, "(&s)", &status);
		_W("[__COOLDOWN__] status(%s)", status);
		if (g_strcmp0(status, COOLDOWN_STATUS_LIMIT) == 0) {
			cooldown_status = COOLDOWN_LIMIT;
			amd_app_status_foreach_running_info(
					__cooldown_limitaction,
					NULL);
		} else if (g_strcmp0(status, COOLDOWN_STATUS_RELEASE) == 0) {
			cooldown_status = COOLDOWN_RELEASE;
			__cooldown_release();
		} else {
			result = -1;
		}
	} else {
		_E("Invalid request(%s)", method_name);
		result = -1;
	}

end:
	g_dbus_method_invocation_return_value(invocation,
			g_variant_new("(i)", result));
}

static guint __get_owner_id(GDBusConnection *connection,
		const char *interface_name)
{
	GError *err = NULL;
	guint owner_id = 0;
	GVariant* result;

	result = g_dbus_connection_call_sync(connection,
			DBUS_SERVICE_DBUS,
			DBUS_PATH_DBUS,
			DBUS_INTERFACE_DBUS,
			"RequestName",
			g_variant_new("(su)",
				interface_name,
				G_BUS_NAME_OWNER_FLAGS_NONE),
			G_VARIANT_TYPE("(u)"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&err);
	if (err) {
		_E("g_dbus_connection_call_sync() is failed. error(%s)",
				err->message);
		g_error_free(err);
		return 0;
	}

	if (result == NULL) {
		_E("Failed to get RequestName");
		return 0;
	}

	g_variant_get(result, "(u)", &owner_id);
	if (owner_id == 0) {
		_E("Acquiring the own name is failed");
		g_variant_unref(result);
		return 0;
	}

	_D("Acquiring the own name : %u", owner_id);
	g_variant_unref(result);

	return owner_id;
}

static int __init_request_handler(void *user_data)
{
	static const char introspection_xml[] =
		"<node>"
		"  <interface name='org.tizen.application.manager'>"
		"    <method name='Cooldown'>"
		"      <arg type='s' name='status' direction='in'/>"
		"      <arg type='i' name='ret' direction='out'/>"
		"    </method>"
		"  </interface>"
		"</node>";
	static const GDBusInterfaceVTable interface_vtable = {
		__handle_method_call,
		NULL,
		NULL
	};
	GDBusNodeInfo *introspection_data;
	GError *err = NULL;
	guint owner_id;

	__conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &err);
	if (!__conn) {
		_E("g_bus_get_sync() is failed. error(%s)",
				err->message);
		g_error_free(err);
		return -1;
	}

	introspection_data = g_dbus_node_info_new_for_xml(introspection_xml,
			&err);
	if (!introspection_data) {
		_E("g_dbus_node_info_new_for_xml() is failed. error(%s)",
				err->message);
		g_error_free(err);
		return -1;
	}

	__registration_id = g_dbus_connection_register_object(
			__conn,
			AMD_DBUS_OBJECT_PATH,
			introspection_data->interfaces[0],
			&interface_vtable,
			NULL,
			NULL,
			&err);
	g_dbus_node_info_unref(introspection_data);
	if (__registration_id == 0) {
		_E("g_dbus_connection_register_object() is failed. error(%s)",
				err->message);
		g_error_free(err);
		return -1;
	}

	owner_id = __get_owner_id(__conn, AMD_DBUS_INTERFACE);
	_I("registration_id(%u), owner_id(%u)", __registration_id, owner_id);

	return 0;
}

static void __fini_request_handler(void)
{
	if (__registration_id > 0) {
		g_dbus_connection_unregister_object(__conn, __registration_id);
		__registration_id = 0;
	}

	if (__conn) {
		g_object_unref(__conn);
		__conn = NULL;
	}
}

EXPORT int AMD_MOD_INIT(void)
{
	_D("cooldown init");
	amd_signal_add_ready_cb(__init_request_handler, NULL);

	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_PREPARE_START,
			__on_check_mode);
	amd_noti_listen(AMD_NOTI_MSG_SIGNAL_SEND_WATCHDOG_START,
			__on_check_status);

	return 0;
}

EXPORT void AMD_MOD_FINI(void)
{
	_D("cooldown fini");
	__fini_request_handler();
}
