/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <amd.h>

#include "rua_image_monitor.h"
#include "private.h"

#define PATH_RUN        "/run"
#define FILE_E_IMG      ".e-img"
#define PATH_RUN_E_IMG  PATH_RUN "/" FILE_E_IMG

struct rua_image_monitor_s {
	amd_inotify_watch_info_h directory_create;
	amd_inotify_watch_info_h image_create;
	amd_inotify_watch_info_h image_close_write;
	rua_image_monitor_event_cb callback;
	void *user_data;
};

static struct rua_image_monitor_s *__get_monitor(void)
{
	static struct rua_image_monitor_s inst;

	return &inst;
}

void _rua_image_monitor_clear_all(void)
{
	DIR *dp;
	struct dirent *dentry = NULL;
	char buf[PATH_MAX];
	struct stat statbuf;
	int r;

	dp = opendir(PATH_RUN_E_IMG);
	if (dp == NULL)
		return;

	while ((dentry = readdir(dp))) {
		if (!strcmp(dentry->d_name, ".") ||
				!strcmp(dentry->d_name, ".."))
			continue;

		snprintf(buf, sizeof(buf), "%s/%s",
				PATH_RUN_E_IMG, dentry->d_name);
		r = stat(buf, &statbuf);
		if (r == 0) {
			if (S_ISREG(statbuf.st_mode))
				unlink(buf);
		}
	}

	closedir(dp);
}

void _rua_image_monitor_clear(const char *image)
{
	if (!image)
		return;

	unlink(image);
}

static void __foreach_images(void)
{
	struct rua_image_monitor_s *monitor = __get_monitor();
	DIR *dp;
	struct dirent *dentry;
	char buf[PATH_MAX];
	struct stat statbuf;
	int r;

	if (!monitor->callback)
		return;

	dp = opendir(PATH_RUN_E_IMG);
	if (dp == NULL)
		return;

	while ((dentry = readdir(dp))) {
		if (!strcmp(dentry->d_name, ".") ||
				!strcmp(dentry->d_name, ".."))
			continue;

		snprintf(buf, sizeof(buf), "%s/%s",
				PATH_RUN_E_IMG, dentry->d_name);
		r = stat(buf, &statbuf);
		if (r == 0) {
			if (S_ISREG(statbuf.st_mode))
				monitor->callback(buf, monitor->user_data);
		}
	}
	closedir(dp);
}

static bool __on_image_create(const char *event_name, void *user_data)
{
	if (event_name)
		_D("%s created", event_name);

	return true;
}

static bool __on_image_close_write(const char *event_name,
		void *user_data)
{
	struct rua_image_monitor_s *monitor = user_data;
	char path[PATH_MAX];

	if (event_name) {
		snprintf(path, sizeof(path), "%s/%s",
				PATH_RUN_E_IMG, event_name);
		if (monitor->callback)
			monitor->callback(path, monitor->user_data);

		_D("%s opened for writing was closed ", event_name);
	}

	return true;
}

static void __fini_rua_image_monitor(void)
{
	struct rua_image_monitor_s *monitor = __get_monitor();

	if (monitor->image_close_write) {
		amd_inotify_rm_watch(monitor->image_close_write);
		monitor->image_close_write = NULL;
	}

	if (monitor->image_create) {
		amd_inotify_rm_watch(monitor->image_create);
		monitor->image_create = NULL;
	}
}

static int __init_rua_image_monitor(void)
{
	struct rua_image_monitor_s *monitor = __get_monitor();

	monitor->image_create = amd_inotify_add_watch(PATH_RUN_E_IMG,
			IN_CREATE, __on_image_create, monitor);
	if (!monitor->image_create)
		_W("Failed to add inotify watch %s", PATH_RUN_E_IMG);

	monitor->image_close_write = amd_inotify_add_watch(PATH_RUN_E_IMG,
			IN_CLOSE_WRITE, __on_image_close_write, monitor);
	if (!monitor->image_close_write) {
		_E("Failed to add inotify watch %s", PATH_RUN_E_IMG);
		return -1;
	}

	__foreach_images();

	return 0;
}

static bool __on_directory_create(const char *event_name, void *user_data)
{
	struct rua_image_monitor_s *monitor = user_data;

	if (!event_name) {
		_E("Invalid parameter");
		return true;
	}

	if (!strcmp(event_name, FILE_E_IMG)) {
		__init_rua_image_monitor();
		monitor->directory_create = NULL;
		return false;
	}

	return true;
}

int _rua_image_monitor_set_event_cb(rua_image_monitor_event_cb callback,
		void *user_data)
{
	struct rua_image_monitor_s *monitor = __get_monitor();

	if (!callback) {
		_E("Invalid parameter");
		return -1;
	}

	monitor->callback = callback;
	monitor->user_data = user_data;

	return 0;
}

int _rua_image_monitor_init(void)
{
	struct rua_image_monitor_s *monitor = __get_monitor();
	int ret;

	_D("RUA Image Monitor Initialize");

	ret = access(PATH_RUN_E_IMG, F_OK);
	if (ret < 0) {
		_W("Failed to access %s", PATH_RUN_E_IMG);
		monitor->directory_create = amd_inotify_add_watch(PATH_RUN,
				IN_CREATE, __on_directory_create, monitor);
		if (!monitor->directory_create) {
			_E("Failed to add inotify watch %s", PATH_RUN);
			return -1;
		}

		return 0;
	}

	ret = __init_rua_image_monitor();
	if (ret < 0)
		return ret;

	return 0;
}

void _rua_image_monitor_fini(void)
{
	struct rua_image_monitor_s *monitor = __get_monitor();

	_D("RUA Image Monitor Finalize");

	__fini_rua_image_monitor();

	if (monitor->directory_create) {
		amd_inotify_rm_watch(monitor->directory_create);
		monitor->directory_create = NULL;
	}
}
