/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <glib.h>
#include <gio/gio.h>
#include <aul.h>
#include <aul_sock.h>
#include <bundle.h>
#include <bundle_internal.h>
#include <rua_internal.h>
#include <rua_stat_internal.h>
#include <aul.h>
#include <amd.h>

#include "rua_request.h"
#include "rua_info.h"
#include "private.h"

static int __dispatch_update_rua_stat(amd_request_h req)
{
	char *caller = NULL;
	char *tag = NULL;
	bundle *kb;
	uid_t uid;
	int ret;

	kb = amd_request_get_bundle(req);
	if (!kb) {
		_E("Invalid request");
		amd_request_send_result(req, -EINVAL);
		return -1;
	}

	bundle_get_str(kb, AUL_SVC_K_RUA_STAT_CALLER, &caller);
	bundle_get_str(kb, AUL_SVC_K_RUA_STAT_TAG, &tag);

	uid = amd_request_get_target_uid(req);
	ret = rua_stat_usr_db_update(caller, tag, uid);
	amd_request_send_result(req, ret);
	_D("rua_stat_usr_db_update - uid(%d), result(%d)", uid, ret);

	return 0;
}

static int __dispatch_add_history(amd_request_h req)
{
	struct rua_rec rec = {0,};
	char *time_str = NULL;
	bundle *kb;
	uid_t uid;
	int ret;

	kb = amd_request_get_bundle(req);
	if (!kb) {
		_E("Invalid request");
		amd_request_send_result(req, -EINVAL);
		return -1;
	}

	bundle_get_str(kb, AUL_K_RUA_PKGNAME, &rec.pkg_name);
	bundle_get_str(kb, AUL_K_RUA_APPPATH, &rec.app_path);
	bundle_get_str(kb, AUL_K_RUA_ARG, &rec.arg);
	bundle_get_str(kb, AUL_K_RUA_TIME, &time_str);
	if (time_str != NULL)
		rec.launch_time = atoi(time_str);
	else
		rec.launch_time = (int)time(NULL);
	bundle_get_str(kb, AUL_K_RUA_INSTANCE_ID, &rec.instance_id);
	bundle_get_str(kb, AUL_K_RUA_INSTANCE_NAME, &rec.instance_name);
	bundle_get_str(kb, AUL_K_RUA_ICON, &rec.icon);
	bundle_get_str(kb, AUL_K_RUA_URI, &rec.uri);

	uid = amd_request_get_target_uid(req);
	ret = rua_usr_db_add_history(&rec, uid);
	amd_request_send_result(req, ret);
	LOGD("rua_usr_db_add_history - uid(%d), result(%d)", uid, ret);

	return 0;
}

static int __dispatch_remove_history(amd_request_h req)
{
	bundle *kb;
	uid_t uid;
	int ret;

	kb = amd_request_get_bundle(req);
	if (!kb) {
		_E("Invalid request");
		amd_request_send_result(req, -EINVAL);
		return -1;
	}

	uid = amd_request_get_target_uid(req);
	ret = rua_usr_db_delete_history(kb, uid);
	amd_request_send_result(req, ret);

	_rua_info_remove(kb, uid);
	LOGD("rua_usr_db_delete_history - uid(%d), result(%d)", uid, ret);

	return 0;
}

static amd_request_cmd_dispatch __dispatch_table[] = {
	{
		.cmd = APP_UPDATE_RUA_STAT,
		.callback = __dispatch_update_rua_stat
	},
	{
		.cmd = APP_ADD_HISTORY,
		.callback = __dispatch_add_history
	},
	{
		.cmd = APP_REMOVE_HISTORY,
		.callback = __dispatch_remove_history
	},

};

static amd_cynara_checker __cynara_checkers[] = {
	{
		.cmd = APP_UPDATE_RUA_STAT,
		.checker = amd_cynara_simple_checker,
		.data = PRIVILEGE_PLATFORM
	},
	{
		.cmd = APP_ADD_HISTORY,
		.checker = amd_cynara_simple_checker,
		.data = PRIVILEGE_PLATFORM
	},
	{
		.cmd = APP_REMOVE_HISTORY,
		.checker = amd_cynara_simple_checker,
		.data = PRIVILEGE_PLATFORM
	},
};

int _rua_request_init(void)
{
	int ret;

	_D("RUA Request Initialize");

	ret = amd_request_register_cmds(__dispatch_table,
			ARRAY_SIZE(__dispatch_table));
	if (ret < 0) {
		_E("Failed to register cmds");
		return -1;
	}

	ret = amd_cynara_register_checkers(__cynara_checkers,
			ARRAY_SIZE(__cynara_checkers));
	if (ret < 0) {
		_E("Failed to register checkers");
		return -1;
	}

	return 0;
}

void _rua_request_fini(void)
{
	_D("RUA Request Finalize");

}
