/*
 * Copyright (c) 2015 - 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <ctype.h>
#include <glib.h>
#include <dlog.h>
#include <aul.h>
#include <aul_svc.h>
#include <bundle_internal.h>
#include <aul_sock.h>
#include <system_info.h>
#include <aul_comp_types.h>

#include <amd.h>

#include "amd_screen_connector.h"
#include "app_group.h"
#include "app_group_wayland.h"
#include "app_group_request.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "AMD_APP_GROUP"

#define APP_SVC_K_LAUNCH_MODE   "__APP_SVC_LAUNCH_MODE__"
#define STATUS_FOREGROUND "fg"
#define STATUS_BACKGROUND "bg"
#define TIZEN_FEATURE_TERMINATE_UNMANAGEABLE_APP \
	(!(amd_config_get_tizen_profile() & (AMD_TIZEN_PROFILE_TV)))

#ifndef AUL_K_ORG_CALLER_INSTANCE_ID
#define AUL_K_ORG_CALLER_INSTANCE_ID "__AUL_ORG_CALLER_INSTANCE_ID__"
#endif

#ifndef AUL_K_CALLER_INSTANCE_ID
#define AUL_K_CALLER_INHSTANCE_ID "__AUL_CALLER_INSTANCE_ID__"
#endif

typedef enum {
	APP_GROUP_LAUNCH_MODE_SINGLE = 0,
	APP_GROUP_LAUNCH_MODE_GROUP,
	APP_GROUP_LAUNCH_MODE_CALLER,
} app_group_launch_mode_e;

struct app_group_context_s {
	char *id;
	char *caller_id;
	pid_t pid;
	pid_t caller_pid;
	int wid;
	int status;
	app_group_launch_mode_e launch_mode;
	bool fg;
	bool group_sig;
	bool can_be_leader;
	bool reroute;
	bool can_shift;
	bool recycle;
};

struct launch_context_s {
	app_group_launch_mode_e mode;
	bool can_attach;
	pid_t leader_pid;
	const char *leader_id;
	bool new_instance;
};

struct app_group_s {
	char *leader_id;
	pid_t leader_pid;
	GList *list;
};

static GList *__app_group_list;
static GList *__recycle_bin;
static struct launch_context_s __launch_context;

static void __prepare_to_suspend(int pid, uid_t uid);

static void __destroy_app_group_context(gpointer data)
{
	struct app_group_context_s *ctx = (struct app_group_context_s *)data;

	amd_noti_send(AMD_NOTI_MSG_APP_GROUP_DESTROY_APP_GROUP_CONTEXT,
			ctx->pid, ctx->wid, ctx->id, NULL);

	if (ctx->caller_id)
		free(ctx->caller_id);
	if (ctx->id)
		free(ctx->id);
	free(ctx);
}

static struct app_group_context_s *__create_app_group_context(
		pid_t pid, const char *id,
		pid_t caller_pid, const char *caller_id,
		app_group_launch_mode_e launch_mode,
		bool can_shift, bool recycle)
{
	struct app_group_context_s *ctx;

	ctx = calloc(1, sizeof(struct app_group_context_s));
	if (!ctx) {
		LOGE("Out of memory");
		return NULL;
	}

	ctx->id = strdup(id);
	if (!ctx->id) {
		LOGE("Failed to duplicate ID(%s)", id);
		free(ctx);
		return NULL;
	}

	ctx->caller_id = strdup(caller_id ? caller_id : "NULL");
	if (!ctx->caller_id) {
		LOGE("Failed to duplicate caller ID(%s)", caller_id);
		__destroy_app_group_context(ctx);
		return NULL;
	}

	ctx->pid = pid;
	ctx->caller_pid = caller_pid;
	ctx->wid = 0;
	ctx->fg = false;
	ctx->can_be_leader = false;
	ctx->reroute = false;
	ctx->launch_mode = launch_mode;
	ctx->can_shift = can_shift;
	ctx->recycle = recycle;

	return ctx;
}

static void __destroy_app_group(gpointer data)
{
	struct app_group_s *group = (struct app_group_s *)data;

	if (group->list)
		g_list_free_full(group->list, __destroy_app_group_context);
	if (group->leader_id)
		free(group->leader_id);
	free(group);
}

static struct app_group_s *__create_app_group(const char *leader_id,
		pid_t leader_pid)
{
	struct app_group_s *group;

	group = calloc(1, sizeof(struct app_group_s));
	if (!group) {
		LOGE("Out of memory");
		return NULL;
	}

	group->leader_id = strdup(leader_id);
	if (!group->leader_id) {
		LOGE("Failed to duplicate leader ID");
		free(group);
		return NULL;
	}

	group->leader_pid = leader_pid;

	return group;
}

static gint __compare_context_id(gconstpointer a, gconstpointer b)
{
	struct app_group_context_s *ctx = (struct app_group_context_s *)a;
	const char *id = (const char *)b;

	return strcmp(ctx->id, id);
}

static gint __compare_group_id(gconstpointer a, gconstpointer b)
{
	app_group_h group = (app_group_h)a;
	const char *id = (const char *)b;

	return strcmp(group->leader_id, id);
}

const char *_app_group_get_id(pid_t pid)
{
	amd_app_status_h app_status;
	int app_type;

	app_status = amd_app_status_find_by_effective_pid(pid);
	if (!app_status) {
		LOGE("Failed to find app status. pid(%d)", pid);
		return NULL;
	}

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type == AMD_AT_COMPONENT_BASED_APP) {
		LOGE("The type of the application(%d) is component-based", pid);
		return NULL;
	}

	return amd_app_status_get_instance_id(app_status);
}

static pid_t __get_caller_pid(bundle *kb)
{
	const char *pid_str;

	pid_str = bundle_get_val(kb, AUL_K_ORG_CALLER_PID);
	if (!pid_str) {
		pid_str = bundle_get_val(kb, AUL_K_CALLER_PID);
		if (!pid_str)
			return -1;
	}

	return atoi(pid_str);
}

static const char *__get_caller_id(bundle *kb)
{
	const char *id;

	id = bundle_get_val(kb, AUL_K_ORG_CALLER_INSTANCE_ID);
	if (!id)
		id = bundle_get_val(kb, AUL_K_CALLER_INSTANCE_ID);

	return id;
}

app_group_node_h _app_group_node_find(const char *id)
{
	app_group_h group;
	app_group_node_h node;
	GList *iter;

	if (!id) {
		LOGE("Invalid parameter");
		return NULL;
	}

	iter = __app_group_list;
	while (iter) {
		group = (app_group_h)iter->data;
		node = g_list_find_custom(group->list, id,
				__compare_context_id);
		if (node)
			return node;

		iter = g_list_next(iter);
	}

	return NULL;
}

app_group_h _app_group_find(const char *id)
{
	app_group_h group;
	app_group_node_h node;
	GList *iter;

	if (!id) {
		LOGE("Invalid parameter");
		return NULL;
	}

	iter = __app_group_list;
	while (iter) {
		group = (app_group_h)iter->data;
		node = g_list_find_custom(group->list, id,
				__compare_context_id);
		if (node)
			return group;

		iter = g_list_next(iter);
	}

	return NULL;
}

pid_t _app_group_get_leader_pid(app_group_h h)
{
	if (!h)
		return -1;

	return h->leader_pid;
}

static struct app_group_context_s *__detach_context_from_recycle_bin(
		const char *id)
{
	struct app_group_context_s *ctx;
	GList *found;

	if (!id)
		return NULL;

	found = g_list_find_custom(__recycle_bin, id, __compare_context_id);
	if (!found)
		return NULL;

	ctx = (struct app_group_context_s *)found->data;
	__recycle_bin = g_list_delete_link(__recycle_bin, found);

	return ctx;
}

int _app_group_node_get_window(app_group_node_h node)
{
	struct app_group_context_s *ctx;

	if (!node)
		return 0;

	ctx = (struct app_group_context_s *)node->data;
	if (ctx)
		return ctx->wid;

	return 0;
}

const char *_app_group_node_get_id(app_group_node_h node)
{
	struct app_group_context_s *ctx;

	if (!node)
		return NULL;

	ctx = (struct app_group_context_s *)node->data;
	if (ctx)
		return ctx->id;

	return NULL;
}

int _app_group_node_set_window(app_group_node_h node, int wid)
{
	struct app_group_context_s *ctx;
	int prev_wid = 0;
	int next_wid = 0;
	int caller_wid;
	app_group_node_h caller_node;
	amd_comp_status_h comp = NULL;

	if (!node) {
		LOGE("Invalid parameter");
		return -1;
	}

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx)
		return -1;

	ctx->wid = wid;
	comp = amd_comp_status_find_by_instance_id(ctx->id);
	if (comp)
		amd_comp_status_set_window(comp, wid);

	prev_wid = _app_group_node_get_window(g_list_previous(node));
	if (prev_wid != 0)
		_app_group_wayland_attach_window(prev_wid, wid);

	if (ctx->can_shift && ctx->caller_id) {
		caller_node = _app_group_node_find(ctx->caller_id);
		caller_wid = _app_group_node_get_window(caller_node);
		if (caller_wid != 0)
			_app_group_wayland_attach_window(caller_wid, wid);
	}

	next_wid = _app_group_node_get_window(g_list_next(node));
	if (next_wid != 0)
		_app_group_wayland_attach_window(wid, next_wid);

	return 0;
}

static bool __app_group_node_is_leader(app_group_node_h node)
{
	struct app_group_context_s *ctx;
	app_group_h group;

	if (!node)
		return false;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx) {
		LOGE("App group context is nullptr");
		return false;
	}

	group = _app_group_find(ctx->id);
	if (group && g_list_first(group->list) == node)
		return true;

	return false;
}

static app_group_node_h __app_group_node_find_second_leader(const char *id)
{
	struct app_group_context_s *ctx;
	app_group_h group;
	GList *node;

	group = _app_group_find(id);
	if (!group)
		return NULL;

	node = g_list_next(group->list);
	if (node) {
		ctx = (struct app_group_context_s *)node->data;
		if (ctx && ctx->can_be_leader) {
			LOGW("Sencond leader. leader_id(%s), id(%s)",
					id, ctx->id);
			return node;
		}
	}

	return NULL;
}

static void __app_group_remove(app_group_node_h node)
{
	struct app_group_context_s *ctx;
	app_group_h group;

	if (!node)
		return;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx) {
		LOGE("App group context is nullptr");
		return;
	}

	group = _app_group_find(ctx->id);
	if (!group) {
		LOGE("Failed to find app group. %s", ctx->id);
		return;
	}

	group->list = g_list_delete_link(group->list, node);
	__destroy_app_group_context(ctx);

	if (g_list_length(group->list) == 0) {
		__app_group_list = g_list_remove(__app_group_list, group);
		__destroy_app_group(group);
	}
}

static pid_t __app_group_node_get_pid(app_group_node_h node)
{
	struct app_group_context_s *ctx;

	if (!node)
		return -1;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx)
		return -1;

	return ctx->pid;
}

static void __app_group_node_remove(app_group_node_h node)
{
	struct app_group_context_s *ctx;
	char id[512];

	if (!node)
		return;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx)
		return;

	snprintf(id, sizeof(id), "%s", ctx->id);

	__app_group_remove(node);
	ctx = __detach_context_from_recycle_bin(id);
	if (ctx)
		free(ctx);
}

static struct app_group_context_s *__detach_context_from_app_group(
		const char *id)
{
	struct app_group_context_s *ctx;
	app_group_h group;
	GList *found;

	group = _app_group_find(id);
	if (!group)
		return NULL;

	found = g_list_find_custom(group->list, id, __compare_context_id);
	if (!found)
		return NULL;

	ctx = (struct app_group_context_s *)found->data;
	group->list = g_list_delete_link(group->list, found);

	if (g_list_length(group->list) == 0) {
		__app_group_list = g_list_remove(__app_group_list, group);
		__destroy_app_group(group);
	}

	return ctx;
}

void _app_group_node_clear_top(app_group_node_h node, uid_t uid)
{
	const char *id;
	int wid;
	pid_t pid;
	GList *iter;
	GList *curr;

	if (!node)
		return;

	iter = g_list_last(node);
	while (iter && iter != node) {
		curr = iter;
		iter = g_list_previous(iter);
		wid = _app_group_node_get_window(curr);
		_app_group_wayland_detach_window(wid);
		pid = __app_group_node_get_pid(curr);
		aul_send_app_terminate_request_signal(pid, NULL, NULL, NULL);
		id = _app_group_node_get_id(curr);
		amd_launch_term_sub_inst(pid, id, uid);
		__app_group_node_remove(curr);
	}
}

static void __app_group_node_remove_leader(const char *id)
{
	struct app_group_context_s *ctx;
	app_group_node_h node;
	app_group_h group;
	GList *found;

	found = g_list_find_custom(__app_group_list, id, __compare_group_id);
	if (!found)
		return;

	group = (app_group_h)found->data;
	if (!group)
		return;

	if (!g_list_next(group->list))
		return;

	node = group->list;
	group->list = NULL;

	ctx = (struct app_group_context_s *)node->data;
	if (ctx)
		__destroy_app_group_context(ctx);

	node = g_list_delete_link(node, node);
	ctx = (struct app_group_context_s *)node->data;

	__app_group_list = g_list_remove(__app_group_list, group);
	__destroy_app_group(group);

	group = __create_app_group(ctx->id, ctx->pid);
	if (group) {
		group->list = node;
		__app_group_list = g_list_append(__app_group_list, group);
	}
}

static const char *__app_group_node_get_next_caller_id(app_group_node_h node)
{
	struct app_group_context_s *ctx;

	if (!node)
		return NULL;

	node = g_list_next(node);
	if (!node)
		return NULL;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx)
		return NULL;

	return ctx->caller_id;
}

static bool __app_group_node_can_reroute(app_group_node_h node)
{
	struct app_group_context_s *ctx;

	if (!node)
		return false;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx)
		return false;

	return ctx->reroute;
}

static void __app_group_node_reroute(app_group_node_h node)
{
	int prev_wid;
	int next_wid;

	prev_wid = _app_group_node_get_window(g_list_previous(node));
	if (!prev_wid)
		return;

	next_wid = _app_group_node_get_window(g_list_next(node));
	if (!next_wid)
		return;

	LOGD("Reroute");
	_app_group_wayland_attach_window(prev_wid, next_wid);
}

static bool __app_group_node_is_sub(app_group_node_h node)
{
	if (!node)
		return false;

	if (g_list_previous(node))
		return true;

	return false;
}

static void __app_group_node_remove_full(app_group_node_h node, uid_t uid)
{
	struct app_group_context_s *ctx;
	const char *caller_id;

	if (!node)
		return;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx) {
		LOGE("App group context is nullptr");
		return;
	}

	if (__app_group_node_is_leader(node)) {
		LOGW("App group leader. %s:%d", ctx->id, ctx->pid);
		if (!__app_group_node_find_second_leader(ctx->id)) {
			_app_group_node_clear_top(node, uid);
			__app_group_node_remove(node);
		} else {
			__app_group_node_remove_leader(ctx->id);
		}
	} else if (__app_group_node_is_sub(node)) {
		LOGW("App group Sub. %s:%d", ctx->id, ctx->pid);
		caller_id = __app_group_node_get_next_caller_id(node);
		if (__app_group_node_can_reroute(node) ||
			(caller_id && strcmp(caller_id, ctx->id) != 0)) {
			LOGW("App group reroute");
			__app_group_node_reroute(node);
		} else {
			LOGW("App group clear top");
			_app_group_node_clear_top(node, uid);
		}
		__app_group_node_remove(node);
	}
}

static void __app_group_remove_full(pid_t pid, uid_t uid)
{
	struct app_group_context_s *ctx;
	app_group_h group;
	app_group_node_h node;
	GList *iter;
	GList *iter_node;

	iter = __app_group_list;
	while (iter) {
		group = (app_group_h)iter->data;
		iter = g_list_next(iter);
		iter_node = g_list_last(group->list);
		while (iter_node) {
			ctx = (struct app_group_context_s *)iter_node->data;
			node = iter_node;
			iter_node = g_list_previous(iter_node);
			if (ctx && ctx->pid == pid)
				__app_group_node_remove_full(node, uid);
		}
	}
}

static struct app_group_context_s *__app_group_add(
		pid_t leader_pid, const char *leader_id,
		pid_t pid, const char *id,
		pid_t caller_pid, const char *caller_id,
		app_group_launch_mode_e launch_mode,
		bool can_shift, bool recycle)
{
	struct app_group_context_s *ctx;
	app_group_h group;
	app_group_node_h node;
	GList *found;

	if (!id || !leader_id) {
		LOGE("Invalid parameter");
		return NULL;
	}

	ctx = __detach_context_from_recycle_bin(id);
	if (!ctx) {
		ctx = __create_app_group_context(pid, id, caller_pid, caller_id,
				launch_mode, can_shift, recycle);
		if (!ctx) {
			LOGE("Failed to create app group context. %s:%d",
					id, pid);
			return NULL;
		}

		if (leader_pid == pid || ctx->recycle)
			ctx->group_sig = true;
		else
			ctx->group_sig = false;
	}

	found = g_list_find_custom(__app_group_list, leader_id,
			__compare_group_id);
	if (!found) {
		group = __create_app_group(leader_id, leader_pid);
		if (!group) {
			LOGE("Failed to create app group. %s:%d",
					leader_id, leader_pid);
			__destroy_app_group_context(ctx);
			return NULL;
		}

		__app_group_list = g_list_append(__app_group_list,
				group);
		found = g_list_last(__app_group_list);
	}

	group = (app_group_h)found->data;

	node = g_list_find_custom(group->list, id, __compare_context_id);
	if (node) {
		LOGE("Already exists. %s:%d", id, pid);
		__destroy_app_group_context(ctx);
		return NULL;
	}

	group->list = g_list_append(group->list, ctx);
	SECURE_LOGD("[__APP_GROUP__] leader(%s:%d), length(%d)",
			leader_id, leader_pid, g_list_length(group->list));

	if (ctx->wid != 0) {
		node = g_list_last(group->list);
		_app_group_node_set_window(node, ctx->wid);
	}

	return ctx;
}

static void __set_hint(struct app_group_context_s *ctx, bundle *b)
{
	const char *str;

	if (!ctx || !b)
		return;

	str = bundle_get_val(b, AUL_SVC_K_CAN_BE_LEADER);
	if (str && !strcmp(str, "true"))
		ctx->can_be_leader = true;
	str = bundle_get_val(b, AUL_SVC_K_REROUTE);
	if (str && !strcmp(str, "true"))
		ctx->reroute = true;
}

static void __app_group_start(pid_t leader_pid, const char *leader_id,
		pid_t pid, const char *id, bundle *b, bool can_attach,
		app_group_launch_mode_e launch_mode)
{
	bool can_shift = false;
	bool recycle = false;
	const char *str;
	const char *caller_id;
	pid_t caller_pid;
	struct app_group_context_s *ctx;

	LOGD("app_group_start");

	caller_pid = __get_caller_pid(b);
	caller_id = __get_caller_id(b);
	if (!caller_id)
		caller_id = _app_group_get_id(caller_pid);

	str = bundle_get_val(b, AUL_SVC_K_SHIFT_WINDOW);
	if (str && !strcmp(str, "true"))
		can_shift = true;

	str = bundle_get_val(b, AUL_SVC_K_RECYCLE);
	if (str && !strcmp(str, "true"))
		recycle = true;

	if (can_attach) {
		ctx = __app_group_add(leader_pid, leader_id, pid, id,
				caller_pid, caller_id, launch_mode,
				false, recycle);
	} else {
		ctx = __app_group_add(pid, id, pid, id,
				caller_pid, caller_id, launch_mode,
				can_shift, false);
	}
	__set_hint(ctx, b);
}

static void __app_group_context_do_recycle(struct app_group_context_s *ctx)
{
	amd_app_status_h app_status;
	const char *appid;
	const char *pkgid;
	const char *comp_type;
	amd_appinfo_h ai;
	uid_t uid;

	app_status = amd_app_status_find_by_pid(ctx->pid);
	uid = amd_app_status_get_uid(app_status);

	if (ctx->fg) {
		appid = amd_app_status_get_appid(app_status);
		ai = amd_appinfo_find(uid, appid);
		pkgid = amd_appinfo_get_value(ai, AMD_AIT_PKGID);
		comp_type = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);

		LOGD("Send BG signal %s:%s", appid, comp_type);
		aul_send_app_status_change_signal(ctx->pid, appid, pkgid,
				STATUS_BACKGROUND, comp_type);
		amd_app_status_find_service_apps(app_status, STATUS_BG,
				__prepare_to_suspend, true);
		ctx->fg = false;
	}

	__recycle_bin = g_list_append(__recycle_bin, ctx);
	amd_noti_send(AMD_NOTI_MSG_APP_GROUP_DO_RECYCLE_END,
			ctx->pid, uid, NULL, NULL);
}

void _app_group_node_lower(app_group_node_h node, bool *exit)
{
	struct app_group_context_s *ctx;

	if (!node || !exit)
		return;

	ctx = (struct app_group_context_s *)node->data;
	if (__app_group_node_is_sub(node)) {
		if (ctx->recycle && ctx->reroute) {
			__app_group_node_reroute(node);
			if (ctx->wid != 0)
				_app_group_wayland_detach_window(ctx->wid);

			ctx = __detach_context_from_app_group(ctx->id);
			__app_group_context_do_recycle(ctx);
			*exit = false;
		} else {
			*exit = true;
		}
		return;
	}

	*exit = false;
	if (ctx->can_shift) {
		_app_group_wayland_detach_window(ctx->wid);
		ctx->can_shift = false;
		_app_group_wayland_lower_window(ctx->wid);
	}
}

static void __app_group_restart(const char *id, bundle *b)
{
	struct app_group_context_s *ctx;
	app_group_node_h node;
	const char *shift_window;
	int caller_wid;
	const char *caller_id;
	app_group_node_h caller_node;

	if (!id || !b) {
		LOGE("Invalid parameter");
		return;
	}

	node = _app_group_node_find(id);
	if (!node)
		return;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx)
		return;

	ctx->caller_pid = __get_caller_pid(b);

	if (ctx->caller_id) {
		free(ctx->caller_id);
		ctx->caller_id = NULL;
	}

	caller_id = __get_caller_id(b);
	if (!caller_id)
		caller_id = _app_group_get_id(ctx->caller_pid);

	ctx->caller_id = strdup(caller_id ? caller_id : "NULL");

	if (ctx->can_shift) {
		if (ctx->wid != 0)
			_app_group_wayland_detach_window(ctx->wid);
		ctx->can_shift = false;
	}

	shift_window = bundle_get_val(b, AUL_SVC_K_SHIFT_WINDOW);
	if (shift_window && !strcmp(shift_window, "true")) {
		ctx->can_shift = true;
		if (ctx->wid != 0) {
			caller_node = _app_group_node_find(ctx->caller_id);
			caller_wid = _app_group_node_get_window(caller_node);
			if (caller_wid != 0) {
				_app_group_wayland_attach_window(caller_wid,
						ctx->wid);
			} else {
				LOGE("Invalid caller window id");
			}
		}
	}
}

static void __app_group_restart_recycle_bin(const char *leader_id,
		const char *id, bundle *b, bool can_attach)
{
	struct app_group_context_s *ctx;
	app_group_h group;
	app_group_node_h node;
	app_group_node_h caller_node;
	const char *caller_id;
	gint caller_pos;
	GList *found;

	if (!id)
		return;

	ctx = __detach_context_from_recycle_bin(id);
	if (!ctx)
		return;

	ctx->caller_pid = __get_caller_pid(b);

	if (ctx->caller_id) {
		free(ctx->caller_id);
		ctx->caller_id = NULL;
	}

	caller_id = __get_caller_id(b);
	if (!caller_id)
		caller_id = _app_group_get_id(ctx->caller_pid);

	ctx->caller_id = strdup(caller_id ? caller_id : "NULL");
	ctx->group_sig = true;

	caller_node = _app_group_node_find(ctx->caller_id);
	if (!caller_node) {
		__app_group_context_do_recycle(ctx);
		return;
	}

	if (can_attach) {
		found = g_list_find_custom(__app_group_list, leader_id,
				__compare_group_id);
		if (!found) {
			__app_group_context_do_recycle(ctx);
			return;
		}

		group = (app_group_h)found->data;
	} else {
		found = g_list_find_custom(__app_group_list, id,
				__compare_group_id);
		if (!found) {
			group = __create_app_group(ctx->id, ctx->pid);
			if (!group) {
				LOGE("Failed to create app group. %s:%d",
						ctx->id, ctx->pid);
				__app_group_context_do_recycle(ctx);
				return;
			}

			__app_group_list = g_list_append(__app_group_list,
					group);
			found = g_list_last(__app_group_list);
		}

		group = (app_group_h)found->data;
	}

	caller_pos = g_list_position(group->list, caller_node);
	group->list = g_list_insert(group->list, ctx, caller_pos + 1);
	SECURE_LOGD("[__APP_GROUP__] leader(%s:%d), length(%d)",
			group->leader_id,
			group->leader_pid,
			g_list_length(group->list));

	if (ctx->wid != 0) {
		node = _app_group_node_find(id);
		_app_group_node_set_window(node, ctx->wid);
	}
}

static void __prepare_to_wake(int pid, uid_t uid)
{
	int ret;
	int dummy = 0;

	ret = aul_sock_send_raw(pid, uid, APP_SUSPEND, (unsigned char *)&dummy,
			sizeof(int), AUL_SOCK_NOREPLY);
	if (ret < 0) {
		LOGE("Failed to send suspend signal. pid(%d), result(%d)",
				pid, ret);
	}

	LOGD("[__APP_SUSPEND__] pid: %d, uid: %d", pid, uid);
}

static void __prepare_to_suspend(int pid, uid_t uid)
{
	int ret;
	int dummy = 0;

	ret = aul_sock_send_raw(pid, uid, APP_WAKE, (unsigned char *)&dummy,
			sizeof(int), AUL_SOCK_NOREPLY);
	if (ret < 0) {
		LOGE("Failed to send wake signal. pid(%d), result(%d)",
				pid, ret);
	}

	LOGD("[__APP_WAKE__] pid: %d, uid: %d", pid, uid);
}

static void __app_group_send_status(pid_t pid, bool flag, bool force, bool self)
{
	amd_app_status_h app_status;
	amd_appinfo_h ai;
	const char *appid;
	const char *pkgid;
	bool bg_allowed;
	uid_t uid;

	app_status = amd_app_status_find_by_effective_pid(pid);
	if (!app_status) {
		LOGE("Failed to find app status. pid(%d)", pid);
		return;
	}

	appid = amd_app_status_get_appid(app_status);
	uid = amd_app_status_get_uid(app_status);

	ai = amd_appinfo_find(uid, appid);
	if (!ai) {
		LOGE("Failed to find appinfo. %s:%d", appid, uid);
		return;
	}

	pkgid = amd_appinfo_get_value(ai, AMD_AIT_PKGID);
	bg_allowed = amd_suspend_is_allowed_background(ai);

	LOGD("Send %s signal %s", flag ? "FG" : "BG", appid);
	aul_send_app_status_change_signal(pid, appid, pkgid,
			flag ? STATUS_FOREGROUND : STATUS_BACKGROUND,
			APP_TYPE_UI);
	if (!bg_allowed) {
		amd_app_status_find_service_apps(app_status,
				flag ? STATUS_VISIBLE : STATUS_BG,
				flag ? __prepare_to_wake : __prepare_to_suspend,
				flag ? false : true);
		if (!flag && force && self) {
			__prepare_to_suspend(pid, uid);
			amd_suspend_add_timer(pid);
		}
	}
}

static void __app_group_set_flag(app_group_h group, pid_t pid, bool flag,
		bool force)
{
	struct app_group_context_s *ctx;
	GList *iter;

	if (!group)
		return;

	iter = group->list;
	while (iter) {
		ctx = (struct app_group_context_s *)iter->data;
		if (ctx && (ctx->fg != flag || force)) {
			__app_group_send_status(ctx->pid, flag, force,
					ctx->pid == pid);
			ctx->fg = flag;
		}
		iter = g_list_next(iter);
	}
}

static bool __app_group_is_visible(app_group_h group)
{
	struct app_group_context_s *ctx;
	GList *iter;

	if (!group)
		return false;

	iter = group->list;
	while (iter) {
		ctx = (struct app_group_context_s *)iter->data;
		if (ctx && ctx->status == STATUS_VISIBLE)
			return true;

		iter = g_list_next(iter);
	}

	return false;
}

static bool __app_group_can_attach_window(bundle *b, const char *appid,
		app_group_launch_mode_e *launch_mode, uid_t uid)
{
	const char *str;
	const char *mode = NULL;
	const char *comp_type;
	const char *comp_id;
	const char *type;
	amd_appinfo_h ai;
	amd_compinfo_h ci;

	ai = amd_appinfo_find(uid, appid);
	if (!ai)
		return false;

	comp_type = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);
	if (comp_type && !strcmp(comp_type, APP_TYPE_UI)) {
		mode = amd_appinfo_get_value(ai, AMD_AIT_LAUNCH_MODE);
	} else if (comp_type && !strcmp(comp_type, APP_TYPE_COMPONENT_BASED)) {
		comp_id = bundle_get_val(b, AUL_K_COMPONENT_ID);
		if (!comp_id)
			return false;

		ci = amd_compinfo_find(uid, comp_id);
		if (!ci)
			return false;

		type = amd_compinfo_get_value(ci, AMD_COMPINFO_TYPE_TYPE);
		if (!type || strcmp(type, "frame") != 0)
			return false;

		mode = amd_compinfo_get_value(ci,
				AMD_COMPINFO_TYPE_LAUNCH_MODE);
	}

	if (mode && !strcmp(mode, "caller"))
		*launch_mode = APP_GROUP_LAUNCH_MODE_CALLER;
	else if (mode && !strcmp(mode, "group"))
		*launch_mode = APP_GROUP_LAUNCH_MODE_GROUP;
	else
		*launch_mode = APP_GROUP_LAUNCH_MODE_SINGLE;

	switch (*launch_mode) {
	case APP_GROUP_LAUNCH_MODE_CALLER:
		LOGD("Launch mode is caller");
		str = bundle_get_val(b, APP_SVC_K_LAUNCH_MODE);
		if (str && !strcmp(str, "group"))
			return true;
		break;
	case APP_GROUP_LAUNCH_MODE_GROUP:
		return true;
	default:
		return false;
	}

	return false;
}

static bool __app_group_can_be_leader(bundle *b)
{
	const char *str;

	str = bundle_get_val(b, AUL_SVC_K_CAN_BE_LEADER);
	if (str && !strcmp(str, "true"))
		return true;

	return false;
}

static void __app_group_context_remove_from_recycle_bin(pid_t pid)
{
	struct app_group_context_s *ctx;
	GList *iter;

	iter = __recycle_bin;
	while (iter) {
		ctx = (struct app_group_context_s *)iter->data;
		iter = g_list_next(iter);
		if (ctx && ctx->pid == pid) {
			__recycle_bin = g_list_remove(__recycle_bin, ctx);
			__destroy_app_group_context(ctx);
		}
	}
}

static bool __app_group_is_group_mode(bundle *kb, uid_t uid)
{
	const char *str;
	const char *mode = NULL;
	const char *appid;
	const char *comp_type;
	const char *comp_id;
	const char *type;
	amd_appinfo_h ai;
	amd_compinfo_h ci;

	if (!kb)
		return false;

	appid = bundle_get_val(kb, AUL_K_APPID);
	if (!appid)
		return false;

	ai = amd_appinfo_find(uid, appid);
	if (!ai)
		return false;

	comp_type = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);
	if (comp_type && !strcmp(comp_type, APP_TYPE_UI)) {
		mode = amd_appinfo_get_value(ai, AMD_AIT_LAUNCH_MODE);
	} else if (comp_type && !strcmp(comp_type, APP_TYPE_COMPONENT_BASED)) {
		comp_id = bundle_get_val(kb, AUL_K_COMPONENT_ID);
		if (!comp_id)
			return false;

		ci = amd_compinfo_find(uid, comp_id);
		if (!ci)
			return false;

		type = amd_compinfo_get_value(ci, AMD_COMPINFO_TYPE_TYPE);
		if (!type || strcmp(type, "frame") != 0)
			return false;

		mode = amd_compinfo_get_value(ci,
				AMD_COMPINFO_TYPE_LAUNCH_MODE);
	}

	if (mode && !strcmp(mode, "caller")) {
		str = bundle_get_val(kb, APP_SVC_K_LAUNCH_MODE);
		if (str && !strcmp(str, "group"))
			return true;
	} else if (mode && !strcmp(mode, "group")) {
		return true;
	}

	return false;
}

void _app_group_get_leader_pids(int *cnt, pid_t **pids)
{
	app_group_h group;
	GList *iter;
	pid_t *leader_pids;
	int size;
	int idx = 0;


	size = g_list_length(__app_group_list);
	if (size > 0) {
		leader_pids = (pid_t *)calloc(size, sizeof(pid_t));
		if (!leader_pids) {
			LOGE("Out of memory");
			return;
		}

		iter = __app_group_list;
		while (iter) {
			group = (app_group_h)iter->data;
			if (group)
				leader_pids[idx++] = group->leader_pid;

			iter = g_list_next(iter);
		}

		*cnt = size;
		*pids = leader_pids;
	}
}

char **_app_group_get_leader_ids(int *cnt)
{
	app_group_h group;
	GList *iter;
	char **leader_ids;
	int idx = 0;
	int size;

	size = g_list_length(__app_group_list);
	if (size == 0)
		return NULL;

	leader_ids = (char **)calloc(size + 1, sizeof(char *));
	if (!leader_ids) {
		LOGE("Out of memory");
		return NULL;
	}

	iter = __app_group_list;
	while (iter) {
		group = (app_group_h)iter->data;
		if (group && group->leader_id)
			leader_ids[idx++] = strdup(group->leader_id);

		iter = g_list_next(iter);
	}

	*cnt = size;

	return leader_ids;
}

int _app_group_get_count(app_group_h group)
{
	if (!group)
		return -1;

	return g_list_length(group->list);
}

int _app_group_foreach_context(app_group_h group,
		app_group_foreach_context_cb callback,
		void *user_data)
{
	struct app_group_context_s *ctx;
	GList *iter;

	if (!group)
		return -1;

	iter = group->list;
	while (iter) {
		ctx = (struct app_group_context_s *)iter->data;
		if (ctx) {
			callback(ctx->id, ctx->pid, ctx->wid, ctx->fg,
					user_data);
		}
		iter = g_list_next(iter);
	}

	return 0;
}

void _app_group_get_group_pids(app_group_h group, int *cnt, pid_t **pids)
{
	struct app_group_context_s *ctx;
	GList *iter;
	int size;
	pid_t *group_pids;
	int idx = 0;

	if (!group)
		return;

	size = g_list_length(group->list);
	if (size == 0)
		return;

	group_pids = calloc(size, sizeof(pid_t));
	if (!group_pids) {
		LOGE("Out of memory");
		return;
	}

	iter = group->list;
	while (iter) {
		ctx = (struct app_group_context_s *)iter->data;
		if (ctx)
			group_pids[idx++] = ctx->pid;
		iter = g_list_next(iter);
	}

	*cnt = size;
	*pids = group_pids;
}

static void __app_group_context_set_status(struct app_group_context_s *ctx,
		struct app_group_context_s *last_ctx,
		pid_t leader_pid, pid_t pid,
		int status, bool force)
{
	amd_app_status_h app_status;
	const char *pkgid;
	app_group_h group;

	if (status > 0)
		ctx->status = status;

	if ((last_ctx && last_ctx->wid != 0) ||
			status == STATUS_VISIBLE ||
			force == true) {
		group = _app_group_find(ctx->id);
		if (__app_group_is_visible(group)) {
			__app_group_set_flag(group, pid, true, force);
			if (!ctx->group_sig && leader_pid != pid) {
				app_status = amd_app_status_find_by_pid(pid);
				pkgid = amd_app_status_get_pkgid(app_status);
				LOGD("Send group signal %d", pid);
				aul_send_app_group_signal(leader_pid, pid,
						pkgid);
				ctx->group_sig = true;
			}
		} else {
			__app_group_set_flag(group, pid, false, force);
		}
	}
}

static int __app_group_node_set_status(app_group_node_h node, int status,
		bool force)
{
	struct app_group_context_s *ctx;
	struct app_group_context_s *last_ctx = NULL;
	struct app_group_context_s *first_ctx = NULL;
	pid_t leader_pid = -1;
	pid_t pid;

	if (!node)
		return -1;

	ctx = (struct app_group_context_s *)node->data;
	pid = ctx->pid;

	node = g_list_last(node);
	if (node)
		last_ctx = (struct app_group_context_s *)node->data;

	node = g_list_first(node);
	if (node) {
		first_ctx = (struct app_group_context_s *)node->data;
		leader_pid = first_ctx->pid;
	}

	__app_group_context_set_status(ctx, last_ctx, leader_pid, pid,
			status, force);

	return 0;
}

bool _app_group_node_get_fg_flag(app_group_node_h node)
{
	struct app_group_context_s *ctx;

	if (!node)
		return false;

	ctx = (struct app_group_context_s *)node->data;
	if (!ctx)
		return false;

	return ctx->fg;
}

static int __app_group_can_start(const char *appid, bundle *b,
		bool *can_attach, pid_t *leader_pid,
		app_group_launch_mode_e *mode, uid_t uid)
{
	pid_t caller_pid;
	const char *caller_id;
	const char *leader_id = NULL;
	pid_t caller_wid;
	app_group_node_h node;
	app_group_h group;
	int ret = 0;

	if (__app_group_can_attach_window(b, appid, mode, uid)) {
		*can_attach = true;
		caller_pid = __get_caller_pid(b);
		caller_id = __get_caller_id(b);
		if (!caller_id) {
			caller_id = _app_group_get_id(caller_pid);
			if (!caller_id) {
				LOGE("Failed to get caller id. %d", caller_pid);
				ret = -1;
				goto end;
			}
		}

		group = _app_group_find(caller_id);
		*leader_pid = _app_group_get_leader_pid(group);
		if (*leader_pid != -1) {
			leader_id = group->leader_id;
			node = _app_group_node_find(caller_id);
			caller_wid = _app_group_node_get_window(node);
			if (caller_wid == 0) {
				LOGW("Caller(%d) window isn't ready",
						caller_pid);
				if (__app_group_can_be_leader(b))
					*can_attach = false;
				else
					*can_attach = true;
			}
		} else {
			LOGE("Cannot find leader pid");
			if (__app_group_can_be_leader(b)) {
				*can_attach = false;
			} else {
				ret = -1;
				goto end;
			}
		}
	}

end:
	__launch_context.can_attach = *can_attach;
	__launch_context.mode = *mode;
	__launch_context.leader_pid = *leader_pid;
	__launch_context.leader_id = leader_id;

	LOGW("[__APP_GROUP__] can_attach(%d), mode(%d), leader(%s:%d)",
			__launch_context.can_attach,
			__launch_context.mode,
			__launch_context.leader_id,
			__launch_context.leader_pid);

	return ret;
}

static int __app_group_find_pid_from_recycle_bin(const char *appid)
{
	struct app_group_context_s *ctx;
	amd_app_status_h app_status;
	const char *appid_from_bin;
	GList *iter;

	if (!appid)
		return -1;

	iter = __recycle_bin;
	while (iter) {
		ctx = (struct app_group_context_s *)iter->data;
		if (ctx) {
			app_status = amd_app_status_find_by_pid(ctx->pid);
			appid_from_bin = amd_app_status_get_appid(app_status);
			if (appid_from_bin && !strcmp(appid_from_bin, appid))
				return ctx->pid;
		}

		iter = g_list_next(iter);
	}

	return -1;
}

void _app_group_get_idle_pids(int *cnt, pid_t **pids)
{
	struct app_group_context_s *ctx;
	GList *iter;
	pid_t *idle_pids;
	int size;
	int idx = 0;

	size = g_list_length(__recycle_bin);
	if (size <= 0)
		return;

	idle_pids = calloc(size, sizeof(pid_t));
	if (!idle_pids) {
		LOGE("Out of memory");
		return;
	}

	iter = __recycle_bin;
	while (iter) {
		ctx = (struct app_group_context_s *)iter->data;
		if (ctx)
			idle_pids[idx++] = ctx->pid;
		iter = g_list_next(iter);
	}

	*cnt = size;
	*pids = idle_pids;
}

int _app_group_get_idle_count(void)
{
	return g_list_length(__recycle_bin);
}

int _app_group_foreach_idle_context(app_group_foreach_context_cb callback,
		void *user_data)
{
	struct app_group_context_s *ctx;
	GList *iter;

	iter = __recycle_bin;
	while (iter) {
		ctx = (struct app_group_context_s *)iter->data;
		if (ctx) {
			callback(ctx->id, ctx->pid, ctx->wid, ctx->fg,
					user_data);
		}
		iter = g_list_next(iter);
	}

	return 0;
}

static int __on_app_status_update_status_end(const char *msg, int arg1,
		int arg2, void *arg3, bundle *data)
{
	bool force = (bool)arg1;
	bool update_group_info = (bool)arg2;
	amd_app_status_h app_status = (amd_app_status_h)arg3;
	app_group_node_h node;
	pid_t pid;
	int status;
	int app_type;

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type != AMD_AT_UI_APP)
		return AMD_NOTI_CONTINUE;

	status = amd_app_status_get_status(app_status);
	if (update_group_info && status != STATUS_DYING) {
		pid = amd_app_status_get_pid(app_status);
		node = _app_group_node_find(_app_group_get_id(pid));
		__app_group_node_set_status(node, status, force);
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_term_app_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	app_group_node_h node;
	app_group_h group;
	const char *id;
	pid_t pid = (pid_t)arg1;
	amd_request_h req = (amd_request_h)arg3;
	struct app_group_context_s *ctx;
	amd_app_status_h app_status;
	GList *iter;
	uid_t uid;

	app_status = amd_app_status_find_by_effective_pid(pid);
	if (!app_status)
		return AMD_NOTI_CONTINUE;

	if (amd_app_status_get_app_type(app_status) != AMD_AT_UI_APP)
		return AMD_NOTI_CONTINUE;

	id = _app_group_get_id(pid);
	node = _app_group_node_find(id);
	if (__app_group_node_is_leader(node)) {
		group = _app_group_find(id);
		if (!group || !group->list)
			return AMD_NOTI_CONTINUE;

		uid = amd_request_get_target_uid(req);
		iter = g_list_last(group->list);
		while (iter) {
			ctx = (struct app_group_context_s *)iter->data;
			node = iter;
			iter = g_list_previous(iter);
			if (ctx) {
				if (ctx->pid != group->leader_pid) {
					amd_launch_term_sub_inst(ctx->pid,
							ctx->id, uid);
				}
				__app_group_node_remove(node);
			}
		}
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_term_bgapp_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	struct app_group_context_s *ctx;
	amd_app_status_h app_status;
	app_group_node_h node;
	app_group_h group;
	pid_t pid = (pid_t)arg1;
	amd_request_h req = (amd_request_h)arg3;
	const char *id;
	GList *iter;
	int status;
	uid_t uid;

	app_status = amd_app_status_find_by_effective_pid(pid);
	if (!app_status)
		return AMD_NOTI_CONTINUE;

	if (amd_app_status_get_app_type(app_status) != AMD_AT_UI_APP)
		return AMD_NOTI_CONTINUE;

	id = _app_group_get_id(pid);
	node = _app_group_node_find(id);
	if (__app_group_node_is_leader(node)) {
		group = _app_group_find(id);
		if (!group || !group->list)
			return AMD_NOTI_CONTINUE;

		ctx = (struct app_group_context_s *)g_list_last(group->list);
		if (!ctx)
			return AMD_NOTI_CONTINUE;

		app_status = amd_app_status_find_by_pid(ctx->pid);
		status = amd_app_status_get_status(app_status);
		if (status != STATUS_BG)
			return AMD_NOTI_CONTINUE;

		uid = amd_request_get_target_uid(req);
		iter = g_list_last(group->list);
		while (iter) {
			ctx = (struct app_group_context_s *)iter->data;
			node = iter;
			iter = g_list_previous(iter);
			if (ctx) {
				if (ctx->pid != group->leader_pid) {
					amd_launch_term_sub_inst(ctx->pid,
							ctx->id, uid);
				}
				__app_group_node_remove(node);
			}
		}
	}

	return AMD_NOTI_CONTINUE;
}

static int __app_group_prepare(int status, uid_t uid, amd_launch_context_h h,
		bundle *kb)
{
	app_group_launch_mode_e launch_mode = APP_GROUP_LAUNCH_MODE_SINGLE;
	bool can_attach = false;
	pid_t leader_pid = -1;
	const char *appid;
	pid_t pid;
	int ret;

	appid = amd_launch_context_get_appid(h);
	pid = amd_launch_context_get_pid(h);
	if (pid <= 0 || status == STATUS_DYING) {
		ret = __app_group_can_start(appid, kb, &can_attach,
				&leader_pid, &launch_mode, uid);
		if (ret != 0) {
			LOGE("Can't make group info");
			return -1;
		}

		if (can_attach && leader_pid == -1) {
			LOGE("Can't launch singleton app in the same group");
			return -1;
		}
	}

	if (pid == -1 && can_attach) {
		pid = __app_group_find_pid_from_recycle_bin(appid);
		amd_launch_context_set_pid(h, pid);
	}

	return 0;
}

static int __on_launch_prepare_ui_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	int status = arg1;
	uid_t uid = (uid_t)arg2;
	amd_launch_context_h h = (amd_launch_context_h)arg3;
	bundle *kb = data;
	int ret;

	if (__app_group_is_group_mode(kb, uid)) {
		amd_launch_context_set_pid(h, -1);
		amd_launch_context_set_subapp(h, true);
		amd_launch_context_set_app_status(h, NULL);
	} else {
		if (amd_launch_context_is_new_instance(h))
			amd_launch_context_set_subapp(h, true);
		else
			amd_launch_context_set_subapp(h, false);
	}

	ret = __app_group_prepare(status, uid, h, data);
	if (ret < 0)
		return AMD_NOTI_STOP;

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_prepare_component_based_start(const char *msg,
		int arg1, int arg2, void *arg3, bundle *data)
{
	int status = arg1;
	uid_t uid = (uid_t)arg2;
	amd_launch_context_h h = (amd_launch_context_h)arg3;
	bundle *kb = data;
	amd_comp_status_h comp_status = NULL;
	const char *id;
	int ret;

	id = amd_launch_context_get_instance_id(h);
	if (status != COMP_STATUS_DESTROYED) {
		comp_status = amd_comp_status_find_by_instance_id(id);
		if (comp_status)
			status = amd_comp_status_get_status(comp_status);
		else
			status = COMP_STATUS_DESTROYED;
	}

	if (__app_group_is_group_mode(kb, uid)) {
		amd_launch_context_set_subapp(h, true);
		amd_launch_context_set_comp_status(h, NULL);
		__launch_context.new_instance = true;
	} else {
		if (!comp_status || status == COMP_STATUS_DESTROYED) {
			__launch_context.new_instance = true;
		} else {
			__launch_context.new_instance =
				amd_launch_context_is_new_instance(h);
		}
	}

	ret = __app_group_prepare(status, uid, h, data);
	if (ret < 0)
		return AMD_NOTI_STOP;

	return AMD_NOTI_CONTINUE;
}

static void __terminate_unmanageable_app(amd_app_status_h app_status)
{
	app_group_h group;
	amd_app_status_h leader_status;
	amd_appinfo_h ai;
	const char *taskmanage;
	const char *appid;
	bool bg_allowed;
	GList *iter;
	uid_t uid;

	if (!amd_app_status_is_home_app(app_status))
		return;

	iter = __app_group_list;
	while (iter) {
		group = (app_group_h)iter->data;
		iter = g_list_next(iter);
		if (!group)
			continue;

		leader_status = amd_app_status_find_by_pid(group->leader_pid);
		if (!leader_status)
			continue;

		if (amd_app_status_is_home_app(leader_status))
			continue;

		if (amd_app_status_get_status(leader_status) != STATUS_BG)
			continue;

		appid = amd_app_status_get_appid(leader_status);
		uid = amd_app_status_get_uid(leader_status);

		ai = amd_appinfo_find(uid, appid);
		taskmanage = amd_appinfo_get_value(ai, AMD_AIT_TASKMANAGE);
		bg_allowed = amd_suspend_is_allowed_background(ai);
		if (taskmanage && !strcmp(taskmanage, "false") && !bg_allowed) {
			SECURE_LOGW("Terminate leader app(%s:%d)",
					appid, group->leader_pid);
			aul_send_app_terminate_request_signal(group->leader_pid,
					NULL, NULL, NULL);
			amd_launch_term_sub_inst(group->leader_pid,
					group->leader_id, uid);
		}
	}
}

static int __on_launch_status_fg(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	amd_app_status_h app_status = (amd_app_status_h)arg3;

	if (TIZEN_FEATURE_TERMINATE_UNMANAGEABLE_APP)
		__terminate_unmanageable_app(app_status);

	return AMD_NOTI_CONTINUE;
}

static gint __compare_app_status_for_sorting(gconstpointer p1, gconstpointer p2)
{
	amd_app_status_h app_status1 = (amd_app_status_h)p1;
	amd_app_status_h app_status2 = (amd_app_status_h)p2;
	int app_group_cnt1;
	int app_group_cnt2;
	int *app_group_pids1;
	int *app_group_pids2;
	int fg_cnt1;
	int fg_cnt2;
	int timestamp1;
	int timestamp2;
	const char *id1;
	const char *id2;
	app_group_h app_group1;
	app_group_h app_group2;

	if (amd_app_status_get_app_type(app_status1) != AMD_AT_UI_APP ||
		amd_app_status_get_app_type(app_status2) != AMD_AT_UI_APP)
		return 0;

	timestamp1 = amd_app_status_get_timestamp(app_status1);
	timestamp2 = amd_app_status_get_timestamp(app_status2);
	if (timestamp1 > timestamp2)
		return 1;
	else if (timestamp1 < timestamp2)
		return -1;

	id1 = amd_app_status_get_leader_id(app_status1);
	app_group1 = _app_group_find(id1);
	_app_group_get_group_pids(app_group1, &app_group_cnt1,
			&app_group_pids1);
	id2 = amd_app_status_get_leader_id(app_status2);
	app_group2 = _app_group_find(id2);
	_app_group_get_group_pids(app_group2, &app_group_cnt2,
			&app_group_pids2);
	free(app_group_pids1);
	free(app_group_pids2);

	if (app_group_cnt1 < app_group_cnt2)
		return 1;
	else if (app_group_cnt1 > app_group_cnt2)
		return -1;

	fg_cnt1 = amd_app_status_get_fg_cnt(app_status1);
	fg_cnt2 = amd_app_status_get_fg_cnt(app_status2);
	if (fg_cnt1 > fg_cnt2)
		return 1;
	else if (fg_cnt1 < fg_cnt2)
		return -1;

	return 0;
}

static int __on_app_status_term_bg_apps(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	amd_app_status_term_bg_apps(__compare_app_status_for_sorting);

	return AMD_NOTI_CONTINUE;
}

static int __on_comp_status_notify_exit(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	uid_t uid = (uid_t)arg2;
	const char *id = (const char *)arg3;
	app_group_node_h node;

	node = _app_group_node_find(id);
	if (!node)
		return AMD_NOTI_CONTINUE;

	__app_group_node_remove_full(node, uid);

	return AMD_NOTI_CONTINUE;
}

static int __on_comp_status_update_status_end(const char *msg, int arg1,
		int arg2, void *arg3, bundle *data)
{
	bool update_group_info = (bool)arg2;
	amd_comp_status_h comp_status = (amd_comp_status_h)arg3;
	app_group_node_h node;
	const char *id;
	int status;

	status = amd_comp_status_get_status(comp_status);
	if (update_group_info && status != COMP_STATUS_DESTROYED) {
		id = amd_comp_status_get_instance_id(comp_status);
		node = _app_group_node_find(id);

		if (status == COMP_STATUS_RESUMED)
			status = STATUS_VISIBLE;
		else
			status = STATUS_BG;

		__app_group_node_set_status(node, status, false);
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_app_status_app_register_pid(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	pid_t pid = (pid_t)arg1;
	amd_appinfo_h ai = (amd_appinfo_h)arg3;
	const char *comp_type;
	const char *id;

	comp_type = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);
	if (comp_type && !strcmp(comp_type, APP_TYPE_UI)) {
		id = _app_group_get_id(pid);
		__app_group_start(pid, id, pid, id, data, false,
				APP_GROUP_LAUNCH_MODE_SINGLE);
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_complete_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	bool new_process = (bool)arg2;
	amd_appinfo_h ai = (amd_appinfo_h)arg3;
	const char *comp_type;

	comp_type = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);
	if (comp_type && !strcmp(comp_type, APP_TYPE_UI))
		__launch_context.new_instance = new_process;

	return AMD_NOTI_CONTINUE;
}

static int __on_launch_complete_end(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	pid_t pid = (pid_t)arg1;
	uid_t uid = (uid_t)arg2;
	amd_appinfo_h ai = (amd_appinfo_h)arg3;
	amd_appinfo_h ci;
	const char *comp_type;
	const char *comp_id;
	const char *type;
	const char *id = NULL;

	comp_type = amd_appinfo_get_value(ai, AMD_AIT_COMPTYPE);
	if (comp_type && !strcmp(comp_type, APP_TYPE_UI)) {
		id = _app_group_get_id(pid);
	} else if (comp_type && !strcmp(comp_type, APP_TYPE_COMPONENT_BASED)) {
		comp_id = bundle_get_val(data, AUL_K_COMPONENT_ID);
		if (!comp_id)
			return AMD_NOTI_CONTINUE;

		ci = amd_compinfo_find(uid, comp_id);
		if (!ci)
			return AMD_NOTI_CONTINUE;

		type = amd_compinfo_get_value(ci, AMD_COMPINFO_TYPE_TYPE);
		if (!type || strcmp(type, "frame") != 0)
			return AMD_NOTI_CONTINUE;

		id = bundle_get_val(data, AUL_K_INSTANCE_ID);
	}

	if (!id) {
		LOGE("Failed to get id. pid(%d)", pid);
		return AMD_NOTI_CONTINUE;
	}

	if (__launch_context.new_instance) {
		__app_group_start(__launch_context.leader_pid,
				__launch_context.leader_id,
				pid, id, data,
				__launch_context.can_attach,
				__launch_context.mode);
		SECURE_LOGI("Add app group %s:%d", id, pid);
	} else {
		__app_group_restart(id, data);
		__app_group_restart_recycle_bin(__launch_context.leader_id, id,
				data, __launch_context.can_attach);
		SECURE_LOGI("Restart app group  %s:%d", id, pid);
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_app_status_add(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	amd_app_status_h app_status = (amd_app_status_h)arg3;
	int app_type;

	if (__launch_context.leader_id == NULL)
		return AMD_NOTI_CONTINUE;

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type == AMD_AT_UI_APP) {
		amd_app_status_set_leader_id(app_status,
				__launch_context.leader_id);
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_comp_status_add(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	int ret;
	amd_comp_status_h comp_status = (amd_comp_status_h)arg3;
	amd_comp_type_e comp_type;

	if (__launch_context.leader_id == NULL)
		return AMD_NOTI_CONTINUE;

	comp_type = amd_comp_status_get_comp_type(comp_status);
	if (comp_type == AMD_CT_FRAME_COMP) {
		ret = amd_comp_status_set_leader_id(comp_status,
				__launch_context.leader_id);
		if (ret < 0) {
			LOGE("Fail to set leader id (%s)",
					__launch_context.leader_id);
		}
	}

	return AMD_NOTI_CONTINUE;
}

static int __on_app_status_cleanup(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	amd_app_status_h app_status = (amd_app_status_h)arg3;
	pid_t pid;
	uid_t uid;

	if (!app_status)
		return AMD_NOTI_STOP;

	pid = amd_app_status_get_pid(app_status);
	uid = amd_app_status_get_uid(app_status);

	__app_group_remove_full(pid, uid);
	__app_group_context_remove_from_recycle_bin(pid);

	return AMD_NOTI_CONTINUE;
}

static bool __is_running_instance(bundle *kb, uid_t uid)
{
	amd_app_status_h app_status;
	amd_comp_status_h comp_status;
	const char *instance_id;
	const char *app_id;
	int status;

	app_id = bundle_get_val(kb, AUL_K_APPID);
	instance_id = bundle_get_val(kb, AUL_K_INSTANCE_ID);
	app_status = amd_app_status_find_by_instance_id(app_id,
			instance_id, uid);
	if (app_status) {
		if (amd_app_status_get_status(app_status) != STATUS_DYING)
			return true;

		return false;
	}

	comp_status = amd_comp_status_find_by_instance_id(instance_id);
	if (comp_status) {
		status = amd_comp_status_get_status(comp_status);
		if (status != COMP_STATUS_DESTROYED)
			return true;

		return false;
	}

	return false;
}

static int __on_launch_app_start_start(const char *msg, int arg1, int arg2,
		void *arg3, bundle *data)
{
	amd_request_h req = (amd_request_h)arg3;
	bundle *kb = data;
	const char *instance_id;
	const char *launch_mode;
	uid_t uid;

	instance_id = bundle_get_val(kb, AUL_K_INSTANCE_ID);
	if (!instance_id)
		return AMD_NOTI_CONTINUE;

	launch_mode = bundle_get_val(kb, APP_SVC_K_LAUNCH_MODE);
	if (!launch_mode || !strcmp(launch_mode, "single"))
		return AMD_NOTI_CONTINUE;

	uid = amd_request_get_target_uid(req);
	if (__app_group_is_group_mode(kb, uid)) {
		if (!__is_running_instance(kb, uid)) {
			bundle_del(kb, AUL_K_NEW_INSTANCE);
			bundle_add(kb, AUL_K_NEW_INSTANCE, "true");
		}
	} else {
		LOGW("Wrong request");
		bundle_del(kb, AUL_K_INSTANCE_ID);
	}

	return AMD_NOTI_CONTINUE;
}

int _app_group_init(void)
{
	int r;

	LOGD("app group init");
	r = _app_group_wayland_init();
	if (r < 0) {
		LOGE("Failed to initialize app group wayland");
		return -1;
	}

	r = _app_group_request_init();
	if (r < 0) {
		LOGE("Failed to initialize app group request");
		return -1;
	}

	amd_noti_listen(AMD_NOTI_MSG_APP_STATUS_ADD,
			__on_app_status_add);
	amd_noti_listen(AMD_NOTI_MSG_APP_STATUS_APP_REGISTER_PID,
			__on_app_status_app_register_pid);
	amd_noti_listen(AMD_NOTI_MSG_APP_STATUS_CLEANUP,
			__on_app_status_cleanup);
	amd_noti_listen(AMD_NOTI_MSG_APP_STATUS_UPDATE_STATUS_END,
			__on_app_status_update_status_end);
	amd_noti_listen(AMD_NOTI_MSG_APP_STATUS_TERM_BG_APPS,
			__on_app_status_term_bg_apps);
	amd_noti_listen(AMD_NOTI_MSG_COMP_STATUS_ADD,
			__on_comp_status_add);
	amd_noti_listen(AMD_NOTI_MSG_COMP_STATUS_NOTIFY_EXIT,
			__on_comp_status_notify_exit);
	amd_noti_listen(AMD_NOTI_MSG_COMP_STATUS_UPDATE_STATUS_END,
			__on_comp_status_update_status_end);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_COMPLETE_START,
			__on_launch_complete_start);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_COMPLETE_END,
			__on_launch_complete_end);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_STATUS_FG,
			__on_launch_status_fg);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_TERM_APP_START,
			__on_launch_term_app_start);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_TERM_BGAPP_START,
			__on_launch_term_bgapp_start);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_PREPARE_UI_START,
			__on_launch_prepare_ui_start);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_PREPARE_COMPONENT_BASED_START,
			__on_launch_prepare_component_based_start);
	amd_noti_listen(AMD_NOTI_MSG_LAUNCH_APP_START_START,
			__on_launch_app_start_start);

	return 0;
}

void _app_group_fini(void)
{
	_app_group_request_fini();
	_app_group_wayland_fini();

	if (__app_group_list)
		g_list_free_full(__app_group_list, __destroy_app_group);

	if (__recycle_bin)
		g_list_free_full(__recycle_bin, __destroy_app_group_context);

	LOGD("app group fini");
}
