/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <ctype.h>
#include <glib.h>
#include <dlog.h>
#include <aul.h>
#include <aul_svc.h>
#include <bundle_internal.h>
#include <aul_sock.h>
#include <system_info.h>

#include "amd.h"
#include "amd_screen_connector.h"
#include "app_group.h"
#include "app_group_request.h"
#include "app_group_wayland.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "AMD_APP_GROUP"

#define ARRAY_SIZE(x) (sizeof(x) / sizeof(x[0]))

static app_group_h __find_app_group_by_pid(pid_t pid)
{
	const char *id;

	id = _app_group_get_id(pid);

	return _app_group_find(id);
}

static app_group_node_h __find_app_group_node_by_pid(pid_t pid)
{
	const char *id;

	id = _app_group_get_id(pid);

	return _app_group_node_find(id);
}

static int __get_window_by_pid(pid_t pid)
{
	app_group_node_h node;

	node = __find_app_group_node_by_pid(pid);

	return _app_group_node_get_window(node);
}

static int __get_window_by_appid(const char *appid, uid_t uid)
{
	amd_app_status_h app_status;
	pid_t pid;

	app_status = amd_app_status_find_by_appid(appid, uid);
	if (!app_status) {
		LOGE("Failed to find app status. appid(%s)", appid);
		return 0;
	}

	pid = amd_app_status_get_pid(app_status);

	return __get_window_by_pid(pid);
}

static int __dispatch_app_group_get_window(amd_request_h req)
{
	app_group_node_h node;
	const char *pid_str;
	pid_t pid = -1;
	int wid;
	bundle *b;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		amd_request_send_result(req, -1);
		return -1;
	}

	pid_str = bundle_get_val(b, AUL_K_PID);
	if (pid_str && isdigit(pid_str[0]))
		pid = atoi(pid_str);

	if (pid <= 1) {
		LOGE("Invalid process ID(%d)", pid);
		amd_request_send_result(req, -1);
		return -1;
	}

	node = __find_app_group_node_by_pid(pid);
	if (!node) {
		LOGE("Failed to find app group node. pid(%d)", pid);
		amd_request_send_result(req, -1);
		return -1;
	}

	wid = _app_group_node_get_window(node);
	if (wid <= 0)
		LOGE("Failed to get window");

	amd_request_send_result(req, wid);

	LOGI("pid(%d), wid(%u)", pid, wid);

	return 0;
}

static int __dispatch_app_group_set_window(amd_request_h req)
{
	uid_t uid = amd_request_get_target_uid(req);
	pid_t pid = amd_request_get_pid(req);
	app_group_node_h node;
	const char *wid_str;
	int wid = -1;
	bundle *b;
	int ret;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		return -1;
	}

	wid_str = bundle_get_val(b, AUL_K_WID);
	if (wid_str && isdigit(wid_str[0]))
		wid = atoi(wid_str);

	if (wid <= 0) {
		LOGE("Failed to get window ID");
		return -1;
	}

	node = __find_app_group_node_by_pid(pid);
	if (!node) {
		LOGE("Failed to find app group node. pid(%d)", pid);
		return -1;
	}

	ret = _app_group_node_set_window(node, wid);
	if (ret < 0) {
		LOGE("Failed to set window. pid(%d), wid(%d)",
				pid, wid);
		return -1;
	}

	_screen_connector_add_app_screen(pid, wid, NULL, uid);
	amd_noti_send(AMD_NOTI_MSG_APP_GROUP_WINDOW_SET, pid, wid, NULL, NULL);
	LOGI("pid(%d), wid(%d), result(%d)", pid, wid, ret);

	return ret;
}

static int __dispatch_app_group_get_fg_flag(amd_request_h req)
{
	app_group_node_h node;
	const char *pid_str;
	pid_t pid = -1;
	bundle *b;
	bool fg;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		amd_request_send_result(req, 0);
		return -1;
	}

	pid_str = bundle_get_val(b, AUL_K_PID);
	if (pid_str && isdigit(pid_str[0]))
		pid = atoi(pid_str);

	node = __find_app_group_node_by_pid(pid);
	if (!node) {
		LOGE("Failed to find app group node. pid(%d)", pid);
		amd_request_send_result(req, 0);
		return -1;
	}

	fg = _app_group_node_get_fg_flag(node);
	amd_request_send_result(req, (int)fg);

	LOGI("pid(%d), fg(%s)", pid, fg ? "true" : "false");

	return 0;
}

static int __dispatch_app_group_clear_top(amd_request_h req)
{
	pid_t pid = amd_request_get_pid(req);
	uid_t uid = amd_request_get_target_uid(req);
	app_group_node_h node;

	node = __find_app_group_node_by_pid(pid);
	if (!node) {
		LOGE("Failed to find app group node. pid(%d)", pid);
		amd_request_send_result(req, 0);
		return -1;
	}

	_app_group_node_clear_top(node, uid);
	amd_request_send_result(req, 0);

	LOGI("pid(%d)", pid);

	return 0;
}

static int __dispatch_app_group_get_leader_pid(amd_request_h req)
{
	app_group_h app_group;
	const char *pid_str;
	pid_t leader_pid;
	pid_t pid = -1;
	bundle *b;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		amd_request_send_result(req, -1);
		return -1;
	}

	pid_str = bundle_get_val(b, AUL_K_PID);
	if (pid_str && isdigit(pid_str[0]))
		pid = atoi(pid_str);

	app_group = __find_app_group_by_pid(pid);
	if (!app_group) {
		LOGE("Failed to find app group. pid(%d)", pid);
		amd_request_send_result(req, -1);
		return -1;
	}

	leader_pid = _app_group_get_leader_pid(app_group);
	amd_request_send_result(req, leader_pid);
	LOGI("pid(%d), leader_pid(%d)", pid, leader_pid);

	return 0;
}

static int __dispatch_app_group_get_leader_pids(amd_request_h req)
{
	unsigned char empty[1] = { 0, };
	pid_t *pids = NULL;
	int cnt = 0;

	_app_group_get_leader_pids(&cnt, &pids);

	if (pids == NULL || cnt == 0) {
		amd_request_send_raw(req, APP_GROUP_GET_LEADER_PIDS, empty, 0);
	} else {
		amd_request_send_raw(req, APP_GROUP_GET_LEADER_PIDS,
				(unsigned char *)pids, cnt * sizeof(int));
	}

	if (pids != NULL)
		free(pids);

	LOGI("count(%d)", cnt);

	return 0;
}

static int __dispatch_app_group_get_idle_pids(amd_request_h req)
{
	unsigned char empty[1] = { 0, };
	pid_t *pids = NULL;
	int cnt = 0;

	_app_group_get_idle_pids(&cnt, &pids);
	if (pids == NULL || cnt == 0) {
		amd_request_send_raw(req, APP_GROUP_GET_IDLE_PIDS, empty, 0);
	} else {
		amd_request_send_raw(req, APP_GROUP_GET_IDLE_PIDS,
				(unsigned char *)pids, cnt * sizeof(int));
	}

	if (pids != NULL)
		free(pids);

	LOGI("count(%d)", cnt);

	return 0;
}

static int __dispatch_app_group_get_group_pids(amd_request_h req)
{
	unsigned char empty[1] = { 0 };
	app_group_h app_group;
	pid_t leader_pid = -1;
	pid_t *pids = NULL;
	int cnt = 0;
	char *buf;
	bundle *b;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		amd_request_send_raw(req, APP_GROUP_GET_GROUP_PIDS, empty, 0);
		return -1;
	}

	bundle_get_str(b, AUL_K_LEADER_PID, &buf);
	if (buf && isdigit(buf[0]))
		leader_pid = atoi(buf);

	if (leader_pid <= 1) {
		LOGE("Failed to get leader process ID");
		amd_request_send_raw(req, APP_GROUP_GET_GROUP_PIDS, empty, 0);
		return -1;
	}

	app_group = __find_app_group_by_pid(leader_pid);
	_app_group_get_group_pids(app_group, &cnt, &pids);
	if (pids == NULL || cnt == 0) {
		amd_request_send_raw(req, APP_GROUP_GET_GROUP_PIDS, empty, 0);
	} else {
		amd_request_send_raw(req, APP_GROUP_GET_GROUP_PIDS,
				(unsigned char *)pids, cnt * sizeof(int));
	}

	if (pids != NULL)
		free(pids);

	LOGI("leader_pid(%d), count(%d)", leader_pid, cnt);

	return 0;
}

static int __dispatch_app_group_lower(amd_request_h req)
{
	pid_t pid = amd_request_get_pid(req);
	app_group_node_h node;
	bool exit = false;

	node = __find_app_group_node_by_pid(pid);
	if (!node) {
		LOGE("Failed to find app group node. pid(%d)", pid);
		amd_request_send_result(req, 0);
		return -1;
	}

	_app_group_node_lower(node, &exit);
	amd_request_send_result(req, (int)exit);
	LOGI("pid(%d), exit(%s)", pid, exit ? "true" : "false");

	return 0;
}

static int __app_group_activate_below(pid_t pid, uid_t uid,
		const char *below_appid)
{
	int below_wid;
	int wid;

	if (!below_appid) {
		LOGE("Invalid parameter");
		return -1;
	}

	wid = __get_window_by_pid(pid);
	if (wid == 0) {
		LOGE("Caller(%d) window is not ready", pid);
		return -1;
	}

	below_wid = __get_window_by_appid(below_appid, uid);
	if (below_wid == 0) {
		LOGE("Below app(%s) window is not ready", below_appid);
		return -1;
	}

	_app_group_wayland_activate_below(wid, below_wid);

	return 0;
}

static int __dispatch_app_group_activate_below(amd_request_h req)
{
	const char *appid;
	int ret;
	bundle *b;
	pid_t pid;
	uid_t uid;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		amd_request_send_result(req, -1);
		return -1;
	}

	appid = bundle_get_val(b, AUL_K_APPID);
	if (!appid) {
		LOGE("Failed to get appid");
		amd_request_send_result(req, -1);
		return -1;
	}

	pid = amd_request_get_pid(req);
	uid = amd_request_get_target_uid(req);

	ret = __app_group_activate_below(pid, uid, appid);
	amd_request_send_result(req, ret);
	LOGI("pid(%d), appid(%s), result(%d)", pid, appid, ret);

	return 0;
}

static int __app_group_activate_above(pid_t pid, uid_t uid,
		const char *above_appid)
{
	int above_wid;
	int wid;

	if (!above_appid) {
		LOGE("Invalid parameter");
		return -1;
	}

	wid = __get_window_by_pid(pid);
	if (wid == 0) {
		LOGE("Caller(%d) window is not ready", pid);
		return -1;
	}

	above_wid = __get_window_by_appid(above_appid, uid);
	if (above_wid == 0) {
		LOGE("Above app(%s) window is not ready", above_appid);
		return -1;
	}

	_app_group_wayland_activate_above(wid, above_wid);

	return 0;
}

static int __dispatch_app_group_activate_above(amd_request_h req)
{
	const char *appid;
	int ret;
	bundle *b;
	pid_t pid;
	uid_t uid;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		amd_request_send_result(req, -1);
		return -1;
	}

	appid = bundle_get_val(b, AUL_K_APPID);
	if (!appid) {
		LOGE("Failed to get appid");
		amd_request_send_result(req, -1);
		return -1;
	}

	pid = amd_request_get_pid(req);
	uid = amd_request_get_target_uid(req);

	ret = __app_group_activate_above(pid, uid, appid);
	amd_request_send_result(req, ret);
	LOGI("pid(%d), appid(%s), result(%d)", pid, appid, ret);

	return 0;
}

static int __app_window_attach(const char *parent_appid,
		const char *child_appid, uid_t uid)
{
	int parent_wid;
	int child_wid;

	parent_wid = __get_window_by_appid(parent_appid, uid);
	if (parent_wid == 0) {
		LOGE("Parent(%s) window is not ready", parent_appid);
		return -1;
	}

	child_wid = __get_window_by_appid(child_appid, uid);
	if (child_wid == 0) {
		LOGE("Child(%s) window is not ready", child_appid);
		return -1;
	}

	_app_group_wayland_attach_window(parent_wid, child_wid);

	return 0;
}

static int __dispatch_app_window_attach(amd_request_h req)
{
	uid_t uid = amd_request_get_target_uid(req);
	bundle *b = amd_request_get_bundle(req);
	const char *parent_appid;
	const char *child_appid;
	int ret;

	if (!b) {
		LOGE("Invalid bundle");
		amd_request_send_result(req, -1);
		return -1;
	}

	parent_appid = bundle_get_val(b, AUL_K_PARENT_APPID);
	if (!parent_appid) {
		LOGE("Invalid parameters");
		amd_request_send_result(req, -1);
		return -1;
	}

	child_appid = bundle_get_val(b, AUL_K_CHILD_APPID);
	if (!child_appid) {
		LOGE("Invalid parameters");
		amd_request_send_result(req, -1);
		return -1;
	}

	ret = __app_window_attach(parent_appid, child_appid, uid);
	amd_request_send_result(req, ret);
	LOGI("parent appid(%s), child appid(%s), result(%d)",
			parent_appid, child_appid, ret);

	return 0;
}

static int __app_window_detach(const char *appid, uid_t uid)
{
	int wid;

	wid = __get_window_by_appid(appid, uid);
	if (wid == 0) {
		LOGE("%s window(%u) is invalid", appid, wid);
		return -1;
	}

	_app_group_wayland_detach_window(wid);

	return 0;
}

static int __dispatch_app_window_detach(amd_request_h req)
{
	uid_t uid = amd_request_get_target_uid(req);
	bundle *b = amd_request_get_bundle(req);
	const char *child_appid;
	int ret;

	if (!b) {
		LOGE("Invalid bundle");
		amd_request_send_result(req, -1);
		return -1;
	}

	child_appid = bundle_get_val(b, AUL_K_CHILD_APPID);
	if (!child_appid) {
		LOGE("Invalid parameters");
		amd_request_send_result(req, -1);
		return -1;
	}

	ret = __app_window_detach(child_appid, uid);
	amd_request_send_result(req, ret);
	LOGI("child appid(%s), result(%d)", child_appid, ret);

	return 0;
}

static int __dispatch_app_group_set_window_v2(amd_request_h req)
{
	uid_t uid = amd_request_get_target_uid(req);
	pid_t pid = amd_request_get_pid(req);
	app_group_node_h node;
	const char *id;
	const char *wid_str;
	int wid = 0;
	bundle *b;
	int ret;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		return -1;
	}

	id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	if (!id) {
		LOGE("Failed to get instance ID");
		return -1;
	}

	wid_str = bundle_get_val(b, AUL_K_WID);
	if (wid_str && isdigit(wid_str[0]))
		wid = (uint32_t)atoi(wid_str);

	if (wid <= 0) {
		LOGE("Failed to get window ID");
		return -1;
	}

	node = _app_group_node_find(id);
	if (!node) {
		LOGE("Failed to find app group node. id(%s)", id);
		return -1;
	}

	ret = _app_group_node_set_window(node, wid);
	if (ret < 0) {
		LOGE("Failed to set window. id(%s), wid(%u)",
				id, wid);
		return -1;
	}

	_screen_connector_add_app_screen(pid, wid, id, uid);
	amd_noti_send(AMD_NOTI_MSG_APP_GROUP_WINDOW_SET,
			pid, wid, (void *)id, NULL);
	LOGI("id(%s), wid(%u), result(%d)", id, wid, ret);

	return ret;
}

static int __dispatch_app_group_lower_v2(amd_request_h req)
{
	app_group_node_h node;
	const char *id;
	bool exit = false;
	bundle *b;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		amd_request_send_result(req, 0);
		return -1;
	}

	id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	if (!id) {
		LOGE("Failed to get instance ID");
		amd_request_send_result(req, 0);
		return -1;
	}

	node = _app_group_node_find(id);
	if (!node) {
		LOGE("Failed to find app group node. id(%s)", id);
		amd_request_send_result(req, 0);
		return -1;
	}

	_app_group_node_lower(node, &exit);
	amd_request_send_result(req, (int)exit);
	LOGI("id(%s), exit(%s)", id, exit ? "true" : "false");

	return 0;
}

static int __dispatch_app_group_get_leader_ids(amd_request_h req)
{
	char **leader_ids;
	int cnt = 0;
	bundle *b;
	int ret;
	int i;

	b = bundle_create();
	if (!b) {
		amd_request_send_result(req, -1);
		return -1;
	}

	leader_ids = _app_group_get_leader_ids(&cnt);
	if (!leader_ids) {
		amd_request_send_result(req, -1);
		bundle_free(b);
		return -1;
	}

	bundle_add_str_array(b, AUL_K_LEADER_IDS,
			(const char **)leader_ids, cnt);
	for (i = 0; i < cnt; i++) {
		if (leader_ids[i])
			free(leader_ids[i]);
	}
	free(leader_ids);

	ret = aul_sock_send_bundle_with_fd(amd_request_remove_fd(req),
			APP_GET_INFO_OK, b, AUL_SOCK_NOREPLY);
	bundle_free(b);

	LOGI("count(%d), result(%d)", cnt, ret);

	return 0;
}

static void __foreach_group_info(const char *id, pid_t pid, int wid,
		bool fg, void *user_data)
{
	amd_request_h req = (amd_request_h)user_data;
	int fd = amd_request_get_fd(req);
	amd_app_status_h app_status;
	amd_comp_status_h comp_status;
	int app_type;
	int status;
	char buf[32];
	bundle *b;
	int ret;

	app_status = amd_app_status_find_by_effective_pid(pid);
	if (!app_status) {
		LOGE("Failed to get app status");
		return;
	}

	b = bundle_create();
	if (!b) {
		LOGE("Out of memory");
		return;
	}

	bundle_add(b, AUL_K_INSTANCE_ID, id);
	snprintf(buf, sizeof(buf), "%d", pid);
	bundle_add(b, AUL_K_PID, buf);
	bundle_add(b, AUL_K_APPID, amd_app_status_get_appid(app_status));
	bundle_add(b, AUL_K_PKGID, amd_app_status_get_pkgid(app_status));
	snprintf(buf, sizeof(buf), "%d", (int)wid);
	bundle_add(b, AUL_K_WID, buf);
	snprintf(buf, sizeof(buf), "%d", (int)fg);
	bundle_add(b, AUL_K_FG_FLAG, buf);

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type == AMD_AT_COMPONENT_BASED_APP) {
		comp_status = amd_comp_status_find_by_instance_id(id);
		status = amd_comp_status_get_status(comp_status);
	} else {
		status = amd_app_status_get_status(app_status);
	}
	snprintf(buf, sizeof(buf), "%d", status);
	bundle_add(b, AUL_K_STATUS, buf);

	ret = aul_sock_send_bundle_with_fd(fd, APP_GET_INFO_OK, b,
			AUL_SOCK_ASYNC);
	if (ret < 0)
		LOGE("Failed to send bundle. result(%d)", ret);
	bundle_free(b);
}

static int __dispatch_app_group_get_group_info(amd_request_h req)
{
	app_group_h group;
	int group_count;
	const char *id;
	bundle *b;

	b = amd_request_get_bundle(req);
	if (!b) {
		LOGE("Failed to get bundle");
		amd_request_send_result(req, -1);
		return -1;
	}

	id = bundle_get_val(b, AUL_K_LEADER_ID);
	if (!id) {
		LOGE("Failed to get leader id");
		amd_request_send_result(req, -1);
		return -1;
	}

	group = _app_group_find(id);
	if (!group) {
		SECURE_LOGE("Failed to find app group. id(%s)", id);
		amd_request_send_result(req, -1);
		return -1;
	}

	group_count = _app_group_get_count(group);
	amd_socket_send_result(amd_request_get_fd(req), group_count, false);

	_app_group_foreach_context(group, __foreach_group_info, req);

	LOGI("leader(%s), count(%d)", id, group_count);

	return 0;
}

static int __dispatch_app_group_get_idle_info(amd_request_h req)
{
	int idle_count;

	idle_count = _app_group_get_idle_count();
	amd_socket_send_result(amd_request_get_fd(req), idle_count, false);

	_app_group_foreach_idle_context(__foreach_group_info, req);

	LOGI("count(%d)", idle_count);

	return 0;
}

static amd_request_cmd_dispatch __dispatch_table[] = {
	{
		.cmd = APP_GROUP_GET_WINDOW,
		.callback = __dispatch_app_group_get_window
	},
	{
		.cmd = APP_GROUP_SET_WINDOW,
		.callback = __dispatch_app_group_set_window
	},
	{
		.cmd = APP_GROUP_GET_FG,
		.callback = __dispatch_app_group_get_fg_flag
	},
	{
		.cmd = APP_GROUP_GET_LEADER_PID,
		.callback = __dispatch_app_group_get_leader_pid
	},
	{
		.cmd = APP_GROUP_GET_LEADER_PIDS,
		.callback = __dispatch_app_group_get_leader_pids
	},
	{
		.cmd = APP_GROUP_GET_GROUP_PIDS,
		.callback = __dispatch_app_group_get_group_pids
	},
	{
		.cmd = APP_GROUP_GET_IDLE_PIDS,
		.callback = __dispatch_app_group_get_idle_pids
	},
	{
		.cmd = APP_GROUP_LOWER,
		.callback = __dispatch_app_group_lower
	},
	{	.cmd = APP_GROUP_CLEAR_TOP,
		.callback = __dispatch_app_group_clear_top
	},
	{
		.cmd = APP_GROUP_ACTIVATE_BELOW,
		.callback = __dispatch_app_group_activate_below
	},
	{
		.cmd = APP_GROUP_ACTIVATE_ABOVE,
		.callback = __dispatch_app_group_activate_above
	},
	{
		.cmd = APP_WINDOW_ATTACH,
		.callback = __dispatch_app_window_attach
	},
	{	.cmd = APP_WINDOW_DETACH,
		.callback = __dispatch_app_window_detach
	},
	{
		.cmd = APP_GROUP_SET_WINDOW_V2,
		.callback = __dispatch_app_group_set_window_v2
	},
	{
		.cmd = APP_GROUP_LOWER_V2,
		.callback = __dispatch_app_group_lower_v2
	},
	{
		.cmd = APP_GROUP_GET_LEADER_IDS,
		.callback = __dispatch_app_group_get_leader_ids
	},
	{
		.cmd = APP_GROUP_GET_GROUP_INFO,
		.callback = __dispatch_app_group_get_group_info
	},
	{
		.cmd = APP_GROUP_GET_IDLE_INFO,
		.callback = __dispatch_app_group_get_idle_info
	},
};

static int __verify_instance_cynara_checker(amd_cynara_caller_info_h info,
		amd_request_h req, void *data)
{
	bundle *b = amd_request_get_bundle(req);
	pid_t pid = amd_request_get_pid(req);
	amd_app_status_h app_status;
	amd_comp_status_h comp_status;
	const char *inst_id;
	const char *id;
	int app_type;

	if (!b) {
		LOGE("Failed to get bundle");
		return AMD_CYNARA_RET_DENIED;
	}

	id = bundle_get_val(b, AUL_K_INSTANCE_ID);
	if (!id) {
		LOGE("Failed to get instance ID");
		return AMD_CYNARA_RET_DENIED;
	}

	app_status = amd_app_status_find_by_effective_pid(pid);
	if (!app_status) {
		LOGE("Failed to find app status. pid(%d)", pid);
		return AMD_CYNARA_RET_DENIED;
	}

	app_type = amd_app_status_get_app_type(app_status);
	if (app_type == AMD_AT_COMPONENT_BASED_APP) {
		comp_status = amd_comp_status_find_by_instance_id(id);
		if (!comp_status) {
			LOGE("Failed to find comp status. instance(%s)", id);
			return AMD_CYNARA_RET_DENIED;
		}

		if (amd_comp_status_get_pid(comp_status) != pid) {
			LOGE("Reject request! instance(%s:%d) doesn't exist",
					id, pid);
			return AMD_CYNARA_RET_DENIED;
		}
	} else {
		inst_id = amd_app_status_get_instance_id(app_status);
		if (!inst_id || strcmp(inst_id, id) != 0) {
			LOGE("Reject request! instance(%s:%d) doesn't exist",
					id, pid);
			return AMD_CYNARA_RET_DENIED;
		}
	}

	return AMD_CYNARA_RET_ALLOWED;
}

static amd_cynara_checker __cynara_checkers[] = {
	{
		.cmd = APP_WINDOW_ATTACH,
		.checker = amd_cynara_simple_checker,
		.data = PRIVILEGE_PLATFORM,
		.priority = 10
	},
	{
		.cmd = APP_WINDOW_DETACH,
		.checker = amd_cynara_simple_checker,
		.data = PRIVILEGE_PLATFORM,
		.priority = 10
	},
	{
		.cmd = APP_GROUP_SET_WINDOW_V2,
		.checker = __verify_instance_cynara_checker,
		.data = NULL,
		.priority = 10
	},
	{
		.cmd = APP_GROUP_LOWER_V2,
		.checker = __verify_instance_cynara_checker,
		.data = NULL,
		.priority = 10
	},
	{
		.cmd = APP_GROUP_GET_LEADER_IDS,
		.checker = amd_cynara_simple_checker,
		.data = PRIVILEGE_PLATFORM,
		.priority = 10
	},
	{
		.cmd = APP_GROUP_GET_IDLE_INFO,
		.checker = amd_cynara_simple_checker,
		.data = PRIVILEGE_PLATFORM,
		.priority = 10
	},
};

int _app_group_request_init(void)
{
	int r;

	LOGD("app group request init");

	r = amd_request_register_cmds(__dispatch_table,
			ARRAY_SIZE(__dispatch_table));
	if (r < 0) {
		LOGE("Failed to register cmds");
		return -1;
	}

	r = amd_cynara_register_checkers(__cynara_checkers,
			ARRAY_SIZE(__cynara_checkers));
	if (r < 0) {
		LOGE("Failed to register checkers");
		return -1;
	}

	return 0;
}

void _app_group_request_fini(void)
{
	LOGD("app group request fini");
}
