/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include <amd.h>
#include <dlog.h>
#include <tizen-extension-client-protocol.h>
#include <wayland-client.h>
#include <wayland-tbm-client.h>

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "AMD_APP_GROUP"

struct wayland_context_s {
	struct wl_display *display;
	struct tizen_policy *tizen_policy;
	uint32_t tizen_policy_id;
	bool tizen_policy_initialized;
};

static struct wayland_context_s __context;

int _app_group_wayland_lower_window(int wid)
{
	if (!__context.tizen_policy_initialized) {
		LOGE("tizen_policy is not initialized");
		return -1;
	}

	tizen_policy_lower_by_res_id(__context.tizen_policy, wid);
	wl_display_roundtrip(__context.display);

	return 0;
}

int _app_group_wayland_attach_window(int parent_wid, int child_wid)
{
	if (!__context.tizen_policy_initialized) {
		LOGE("tizen_policy is not initialized");
		return -1;
	}

	tizen_policy_set_transient_for(__context.tizen_policy,
			child_wid, parent_wid);
	wl_display_roundtrip(__context.display);

	return 0;
}

int _app_group_wayland_detach_window(int child_wid)
{
	if (!__context.tizen_policy_initialized) {
		LOGE("tizen_policy is not initialized");
		return -1;
	}

	tizen_policy_unset_transient_for(__context.tizen_policy, child_wid);
	wl_display_roundtrip(__context.display);

	return 0;
}

int _app_group_wayland_activate_below(int wid, int below_wid)
{
	if (!__context.tizen_policy_initialized) {
		LOGE("tizen_policy is not initialized");
		return -1;
	}

	tizen_policy_activate_below_by_res_id(__context.tizen_policy,
			below_wid, wid);
	wl_display_roundtrip(__context.display);

	return 0;
}

int _app_group_wayland_activate_above(int wid, int above_wid)
{
	if (!__context.tizen_policy_initialized) {
		LOGE("tizen_policy is not initialized");
		return -1;
	}

	tizen_policy_activate_above_by_res_id(__context.tizen_policy,
			above_wid, wid);
	wl_display_roundtrip(__context.display);

	return 0;
}

static void __wl_conformant(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface_resource,
		uint32_t is_conformant)
{
}

static void __wl_conformant_area(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface_resource,
		uint32_t conformant_part,
		uint32_t state,
		int32_t x, int32_t y, int32_t w, int32_t h)
{
}

static void __wl_notification_done(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface,
		int32_t level,
		uint32_t state)
{
}

static void __wl_transient_for_done(void *data,
		struct tizen_policy *tizen_policy,
		uint32_t child_id)
{
}

static void __wl_scr_mode_done(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface,
		uint32_t mode,
		uint32_t state)
{
}

static void __wl_iconify_state_changed(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface_resource,
		uint32_t iconified,
		uint32_t force)
{
}

static void __wl_supported_aux_hints(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface_resource,
		struct wl_array *hints,
		uint32_t num_hints)
{
}

static void __wl_allowed_aux_hint(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface_resource,
		int id)
{
}

static void __wl_aux_message(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface_resource,
		const char *key,
		const char *val,
		struct wl_array *options)
{
}

static void __wl_conformant_region(void *data,
		struct tizen_policy *tizen_policy,
		struct wl_surface *surface,
		uint32_t conformant_part,
		uint32_t state,
		int32_t x, int32_t y, int32_t w, int32_t h,
		uint32_t serial)
{
}

static const struct tizen_policy_listener __tizen_policy_listener = {
	__wl_conformant,
	__wl_conformant_area,
	__wl_notification_done,
	__wl_transient_for_done,
	__wl_scr_mode_done,
	__wl_iconify_state_changed,
	__wl_supported_aux_hints,
	__wl_allowed_aux_hint,
	__wl_aux_message,
	__wl_conformant_region,
};

static int __on_wayland_listener_tizen_policy(const char *msg, int arg1,
		int arg2, void *arg3, bundle *data)
{
	uint32_t id = (uint32_t)arg1;
	struct wl_registry *reg = (struct wl_registry *)arg3;

	if (!__context.tizen_policy) {
		__context.display = amd_wayland_get_display();
		__context.tizen_policy_id = id;
		__context.tizen_policy = wl_registry_bind(reg, id,
				&tizen_policy_interface, 7);
		if (__context.tizen_policy) {
			tizen_policy_add_listener(__context.tizen_policy,
					&__tizen_policy_listener,
					__context.display);
			amd_wayland_set_tizen_policy(__context.tizen_policy);
			__context.tizen_policy_initialized = true;
		}
		LOGD("tizen_policy(%p)", __context.tizen_policy);
	}

	return 0;
}

static int __on_wayland_listener_listener_remove(const char *msg, int arg1,
		int arg2, void *arg3, bundle *data)
{
	uint32_t id = (uint32_t)arg1;

	if (__context.tizen_policy_id == id && __context.tizen_policy) {
		tizen_policy_destroy(__context.tizen_policy);
		__context.tizen_policy = NULL;
		__context.tizen_policy_id = 0;
		__context.tizen_policy_initialized = false;
		amd_wayland_set_tizen_policy(__context.tizen_policy);
		LOGW("tizen policy is destroyed");
	}

	return 0;
}

int _app_group_wayland_init(void)
{
	LOGD("app group wayland init");
	amd_noti_listen(AMD_NOTI_MSG_WAYLAND_LISTENER_TIZEN_POLICY,
			__on_wayland_listener_tizen_policy);
	amd_noti_listen(AMD_NOTI_MSG_WAYLAND_LISTENER_REMOVE,
			__on_wayland_listener_listener_remove);

	return 0;
}

void _app_group_wayland_fini(void)
{
	LOGD("app group wayland fini");
}
