/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "amd_logger.h"
#include "amd_util.h"

#define LOG_MAX_SIZE 2500
#define LOG_MAX_STRING_SIZE 128
#define LOG_APPFW_PATH "/var/log/appfw"
#define LOG_PATH "/var/log/appfw/amd"

struct logger_s {
	int fd;
	int index;
};

static gid_t __get_gid(const char *group_name)
{
	char buf[sysconf(_SC_GETGR_R_SIZE_MAX)];
	struct group grp;
	struct group *result;
	int ret;

	ret = getgrnam_r(group_name, &grp, buf, sizeof(buf), &result);
	if (ret < 0) {
		_E("Failed to get group(%s) file entry. errno(%d)",
				group_name, errno);
		return -1;
	}

	return grp.gr_gid;
}

static int __set_ownership(const char *path, uid_t uid, gid_t gid)
{
	int ret;
	int fd;

	fd = open(path, O_RDONLY);
	if (fd < 0) {
		_E("Failed to open directory(%s). errno(%d)", path, errno);
		return -1;
	}

	ret = fchown(fd, uid, gid);
	close(fd);
	if (ret < 0) {
		_E("Failed to change ownership. path(%s), errno(%d)",
				path, errno);
		return -1;
	}

	return 0;
}

int _logger_create(const char *path, logger_h *handle)
{
	struct logger_s *logger;
	off_t offset;

	if (!path || !handle) {
		_E("Invalid parameter");
		return -EINVAL;
	}

	logger = calloc(1, sizeof(struct logger_s));
	if (!logger) {
		_E("Out of memory");
		return -ENOMEM;
	}

	logger->fd = open(path, O_CREAT | O_WRONLY, 0640);
	if (logger->fd < 0) {
		_E("Failed to open path(%s), errno(%d)", path, errno);
		_logger_destroy(logger);
		return -1;
	}

	__set_ownership(path, getuid(), __get_gid("system_share"));

	offset = lseek(logger->fd, 0, SEEK_END);
	if (offset != 0) {
		logger->index = (int)(offset / LOG_MAX_STRING_SIZE);
		if (logger->index >= LOG_MAX_SIZE) {
			logger->index = 0;
			lseek(logger->fd, 0, SEEK_SET);
		}
	}

	*handle = logger;

	return 0;
}

int _logger_destroy(logger_h handle)
{
	if (!handle) {
		_E("Invalid parameter");
		return -EINVAL;
	}

	if (handle->fd > 0)
		close(handle->fd);

	free(handle);

	return 0;
}

int _logger_print(logger_h handle, const char *tag, const char *format, ...)
{
	char format_buf[80];
	char buf[LOG_MAX_STRING_SIZE];
	struct tm tm = { 0, };
	time_t t;
	ssize_t ret;
	va_list ap;
	off_t offset;

	if (!handle || !tag || !format) {
		_E("Invalid parameter");
		return -EINVAL;
	}

	t = time(NULL);
	localtime_r(&t, &tm);

	if (handle->index != 0)
		offset = lseek(handle->fd, 0, SEEK_CUR);
	else
		offset = lseek(handle->fd, 0, SEEK_SET);

	if (offset == -1)
		_E("lseek() is failed. errno(%d)", errno);

	va_start(ap, format);
	vsnprintf(format_buf, sizeof(format_buf), format, ap);
	va_end(ap);

	snprintf(buf, sizeof(buf),
			"[%6d] %04d-%02d-%02d %02d:%02d:%02d %-16s %-80s\n",
			handle->index,
			tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
			tm.tm_hour, tm.tm_min, tm.tm_sec,
			tag, format_buf);
	ret = write(handle->fd, buf, strlen(buf));
	if (ret < 0) {
		_E("Failed to write log message. errno(%d)", errno);
		return -1;
	}

	if (++handle->index >= LOG_MAX_SIZE)
		handle->index = 0;

	return ret;
}

static int __create_directory(const char *path)
{
	gid_t gid;
	int ret;

	ret = access(path, F_OK);
	if (ret == 0)
		return 0;

	ret = mkdir(path, (S_IRUSR | S_IWUSR | S_IXUSR |
				S_IRGRP | S_IWGRP | S_IXGRP));
	if (ret < 0) {
		_E("Failed to create directory(%s). errno(%d)", path, errno);
		return -1;
	}

	gid = __get_gid("system_share");
	if (gid == (gid_t)-1) {
		_E("Failed to get gid");
		return -1;
	}

	return __set_ownership(path, getuid(), gid);
}

int _logger_init(void)
{
	int ret;

	ret = __create_directory(LOG_APPFW_PATH);
	if (ret < 0)
		return ret;

	ret = __create_directory(LOG_PATH);
	if (ret < 0)
		return ret;

	return 0;
}
