/*
 * Copyright (c) 2016 - 2017 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <gio/gio.h>
#include <glib.h>

#include "amd_noti.h"
#include "amd_signal.h"
#include "amd_util.h"
#include "app_signal.h"

#define MAX_LABEL_BUFSZ 1024
#define RETRY_INTERVAL 3

typedef struct {
	int pid;
	void (*callback)(int, int, int, void *);
	void *data;
} proc_status_t;

typedef struct {
	signal_ready_cb callback;
	void *user_data;
} ready_cb_info_t;

static GDBusConnection *system_conn;
static guint startup_finished_sid;
static guint user_session_startup_finished_sid;
static int (*startup_finished_callback)(uid_t, void *);
static void *startup_finished_data;
static uid_t startup_finished_uid;
static bool system_boot_completed;
static bool user_boot_completed;
static guint poweroff_state_sid;
static void (*poweroff_state_callback)(int, void *);
static void *poweroff_state_data;

static GList *__ready_cbs;

static gboolean __dispatch_ready_cb(gpointer data);

static GDBusConnection *__get_system_conn(void)
{
	GError *err = NULL;

	if (system_conn)
		return system_conn;

	system_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &err);
	if (system_conn == NULL) {
		_E("g_bus_get_sync() is failed: %s", err->message);
		g_error_free(err);
		return NULL;
	}

	if (__dispatch_ready_cb(NULL)) {
		g_timeout_add_seconds(RETRY_INTERVAL, __dispatch_ready_cb,
				NULL);
	}

	return system_conn;
}

static int __send_signal(const char *object_path, const char *interface_name,
		const char *signal_name, GVariant *parameters)
{
	GError *err = NULL;
	GDBusConnection *conn;

	conn = __get_system_conn();
	if (conn == NULL)
		return -1;

	if (g_dbus_connection_emit_signal(conn,
					NULL,
					object_path,
					interface_name,
					signal_name,
					parameters,
					&err) == FALSE) {
		_E("g_dbus_connection_emit_signal() is failed: %s",
					err->message);
		g_error_free(err);
		return -1;
	}

	if (g_dbus_connection_flush_sync(conn, NULL, &err) == FALSE) {
		_E("g_dbus_connection_flush_sync() is failed: %s",
				err->message);
		g_error_free(err);
		return -1;
	}

	return 0;
}

int _signal_send_watchdog(int pid, int signal_num)
{
	int r;
	GVariant *param;

	r = _noti_send(NOTI_MSG_SIGNAL_SEND_WATCHDOG_START, pid, signal_num,
			NULL, NULL);
	if (r< 0) {
		_E("Some listeners don't want to continue (pid:%d)", pid);
		return -1;
	}

	param = g_variant_new("(ii)", pid, signal_num);
	if (!param) {
		_E("Out of memory");
		return -1;
	}

	r = __send_signal(RESOURCED_PROC_OBJECT,
			RESOURCED_PROC_INTERFACE,
			RESOURCED_PROC_WATCHDOG_SIGNAL,
			param);
	if (r < 0) {
		_E("Failed to send a watchdog signal - pid(%d)", pid);
		return -1;
	}

	_W("Send a watchdog signal done - pid(%d)", pid);

	return 0;
}

int _signal_send_proc_prelaunch(const char *appid, const char *pkgid,
		int attribute, int category)
{
	int r;
	GVariant *param;

	param = g_variant_new("(ssii)", appid, pkgid, attribute, category);
	if (!param) {
		_E("Out of memory");
		return -1;
	}

	r = __send_signal(RESOURCED_PROC_OBJECT,
			RESOURCED_PROC_INTERFACE,
			RESOURCED_PROC_PRELAUNCH_SIGNAL,
			param);
	if (r < 0) {
		_E("Failed to send a prelaunch signal - appid(%s)", appid);
		return -1;
	}

	_W("send a prelaunch signal done: " \
			"appid(%s) pkgid(%s) attribute(%x) category(%x)",
			appid, pkgid, attribute, category);

	return 0;
}

int _signal_send_tep_mount(char *mnt_path[], const char *pkgid)
{
	GError *err = NULL;
	GDBusMessage *msg;
	GDBusConnection *conn;
	int ret = 0;
	int rv = 0;
	struct stat link_buf = {0,};
	GVariant *param;
	char buf[MAX_LABEL_BUFSZ];

	if (pkgid == NULL) {
		_E("Invalid parameter");
		return -1;
	}

	conn = __get_system_conn();
	if (conn == NULL)
		return -1;

	rv = lstat(mnt_path[0], &link_buf);
	if (rv == 0) {
		rv = unlink(mnt_path[0]);
		if (rv)
			_E("Unable tp remove link file %s", mnt_path[0]);
	}

	msg = g_dbus_message_new_method_call(TEP_BUS_NAME,
					TEP_OBJECT_PATH,
					TEP_INTERFACE_NAME,
					TEP_MOUNT_METHOD);
	if (msg == NULL) {
		_E("g_dbus_message_new_method_call() is failed.");
		return -1;
	}

	snprintf(buf, sizeof(buf), "User::Pkg::%s::RO", pkgid);
	param = g_variant_new("(sss)", mnt_path[0], mnt_path[1], buf);
	g_dbus_message_set_body(msg, param);

	if (g_dbus_connection_send_message(conn,
					msg,
					G_DBUS_SEND_MESSAGE_FLAGS_NONE,
					NULL,
					&err) == FALSE) {
		_E("g_dbus_connection_send_message() is failed: %s",
					err->message);
		ret = -1;
	}

	if (g_dbus_connection_flush_sync(conn, NULL, &err) == FALSE) {
		_E("g_dbus_connection_flush_sync() is failed: %s",
					err->message);
		ret = -1;
	}

	g_object_unref(msg);
	g_clear_error(&err);

	return ret;
}

int _signal_send_tep_unmount(const char *mnt_path)
{
	GError *err = NULL;
	GDBusMessage *msg;
	GDBusConnection *conn;

	conn = __get_system_conn();
	if (conn == NULL)
		return -1;

	msg = g_dbus_message_new_method_call(TEP_BUS_NAME,
					TEP_OBJECT_PATH,
					TEP_INTERFACE_NAME,
					TEP_UNMOUNT_METHOD);
	if (msg == NULL) {
		_E("g_dbus_message_new_method_call() is failed.");
		return -1;
	}

	g_dbus_message_set_body(msg, g_variant_new("(s)", mnt_path));
	if (g_dbus_connection_send_message(conn,
					msg,
					G_DBUS_SEND_MESSAGE_FLAGS_NONE,
					NULL,
					&err) == FALSE) {
		_E("g_dbus_connection_send_message() is failed: %s",
					err->message);
		g_object_unref(msg);
		g_clear_error(&err);
		return -1;
	}

	g_dbus_connection_flush(conn, NULL, NULL, NULL);
	g_object_unref(msg);
	g_clear_error(&err);

	return 0;
}

int _signal_send_proc_suspend(int pid)
{
	GError *err = NULL;
	GDBusConnection *conn;

	conn = __get_system_conn();
	if (conn == NULL)
		return -1;

	if (g_dbus_connection_emit_signal(conn,
					NULL,
					APPFW_SUSPEND_HINT_PATH,
					APPFW_SUSPEND_HINT_INTERFACE,
					APPFW_SUSPEND_HINT_SIGNAL,
					g_variant_new("(i)", pid),
					&err) == FALSE) {
		_E("g_dbus_connection_emit_signal() is failed: %s",
					err->message);
		g_error_free(err);
		return -1;
	}

	if (g_dbus_connection_flush_sync(conn, NULL, &err) == FALSE) {
		_E("g_dbus_connection_flush_sync() is failed: %s",
					err->message);
		g_error_free(err);
		return -1;
	}

	_D("[__SUSPEND__] Send suspend hint, pid: %d", pid);

	return 0;
}

static void __system_bus_signal_handler(GDBusConnection *connection,
		const gchar *sender_name, const gchar *object_path,
		const gchar *interface_name, const char *signal_name,
		GVariant *parameters, gpointer user_data)
{
	guint64 uid = 0;
	int state = -1;

	_W("[SIGNAL_HANDLER] signal(%s)", signal_name);
	if (g_strcmp0(signal_name, SD_STARTUP_FINISHED_SIGNAL) == 0) {
		system_boot_completed = true;
		_D("[SIGNAL_HANDLER] system boot completed");
	} else if (g_strcmp0(signal_name,
			SD_USER_SESSION_STARTUP_FINISHED_SIGNAL) == 0) {
		user_boot_completed = true;
		g_variant_get(parameters, "(t)", &uid);
		startup_finished_uid = (uid_t)uid;
		_D("[SIGNAL_HANDLER] user boot completed");
	} else if (g_strcmp0(signal_name, SYSTEM_POWEROFF_STATE_SIGNAL) == 0) {
		g_variant_get(parameters, "(i)",  &state);
		if (poweroff_state_callback)
			poweroff_state_callback(state, poweroff_state_data);
		_D("[SIGNAL_HANDLER] poweroff state(%d)", state);
	}

	if (system_boot_completed && user_boot_completed) {
		if (startup_finished_callback) {
			startup_finished_callback(startup_finished_uid,
					startup_finished_data);
		}

		user_boot_completed = false;
	}
}

static guint __subscribe_system_bus(const char *object_path,
		const char *interface_name, const char *signal_name)
{
	guint sid;
	GError *err = NULL;
	GDBusConnection *conn;

	conn = __get_system_conn();
	if (conn == NULL)
		return 0;

	sid = g_dbus_connection_signal_subscribe(conn,
						NULL,
						interface_name,
						signal_name,
						object_path,
						NULL,
						G_DBUS_SIGNAL_FLAGS_NONE,
						__system_bus_signal_handler,
						NULL,
						NULL);
	if (sid == 0)
		_E("g_bus_connection_signal_subscribe() is failed");

	g_clear_error(&err);

	return sid;
}

int _signal_subscribe_startup_finished(int (*callback)(uid_t uid, void *data),
		void *user_data)
{
	if (callback == NULL)
		return -1;

	startup_finished_sid = __subscribe_system_bus(SD_OBJECT_PATH,
			SD_MANAGER_INTERFACE,
			SD_STARTUP_FINISHED_SIGNAL);
	if (startup_finished_sid == 0) {
		_E("Failed to subscribe systemd signal");
		return -1;
	}

	user_session_startup_finished_sid = __subscribe_system_bus(
			SD_OBJECT_PATH,
			SD_MANAGER_INTERFACE,
			SD_USER_SESSION_STARTUP_FINISHED_SIGNAL);
	if (user_session_startup_finished_sid == 0) {
		_E("Failed to subscribe systemd signal");
		_signal_unsubscribe_startup_finished();
		return -1;
	}

	startup_finished_callback = callback;
	startup_finished_data = user_data;
	_D("[SIGNAL] subscribe startup finished");

	return 0;
}

int _signal_unsubscribe_startup_finished(void)
{
	GDBusConnection *conn;

	conn = __get_system_conn();
	if (conn == NULL)
		return -1;

	if (!startup_finished_sid && !user_session_startup_finished_sid)
		return 0;

	if (startup_finished_sid) {
		g_dbus_connection_signal_unsubscribe(conn,
				startup_finished_sid);
		startup_finished_sid = 0;
	}

	if (user_session_startup_finished_sid) {
		g_dbus_connection_signal_unsubscribe(conn,
				user_session_startup_finished_sid);
		user_session_startup_finished_sid = 0;
	}

	startup_finished_callback = NULL;
	startup_finished_data = NULL;
	_D("[SIGNAL] unsubscribe startup finished");

	return 0;
}

int _signal_send_display_lock_state(const char *state, const char *flag,
		unsigned int timeout)
{
	GError *err = NULL;
	GDBusConnection *conn;
	GDBusMessage *msg;
	const char *holdkeyblock_string = "holdkeyblock";
	int ret = 0;

	_D("Acquring display lock");
	conn = __get_system_conn();
	if (conn == NULL)
		return -1;

	msg = g_dbus_message_new_method_call(SYSTEM_BUS_NAME,
			SYSTEM_PATH_DISPLAY,
			SYSTEM_INTERFACE_DISPLAY,
			SYSTEM_LOCK_STATE);
	if (msg == NULL) {
		_E("g_dbus_message_new_method_call() is failed");
		return -1;
	}

	g_dbus_message_set_body(msg, g_variant_new("(sssi)", state,
			flag, holdkeyblock_string, timeout));
	if (!g_dbus_connection_send_message(conn, msg,
				G_DBUS_SEND_MESSAGE_FLAGS_NONE, NULL, &err)) {
		_E("Unable to send dbus message for acquring lock as  %s",
				err->message);
		ret = -1;
	}

	_D("Display lock acquired");
	g_object_unref(msg);
	g_dbus_connection_flush_sync(conn, NULL, NULL);
	g_clear_error(&err);
	return ret;
}

int _signal_send_display_unlock_state(const char *state, const char *flag)
{
	GError *err = NULL;
	GDBusConnection *conn;
	GDBusMessage *msg;
	int ret = 0;

	_D("releasing display lock");
	conn = __get_system_conn();
	if (conn == NULL)
		return -1;

	msg = g_dbus_message_new_method_call(SYSTEM_BUS_NAME,
			SYSTEM_PATH_DISPLAY,
			SYSTEM_INTERFACE_DISPLAY,
			SYSTEM_UNLOCK_STATE);
	if (msg == NULL) {
		_E("g_dbus_message_new_method_call() is failed");
		return -1;
	}

	g_dbus_message_set_body(msg, g_variant_new("(ss)", state, flag));
	if (!g_dbus_connection_send_message(conn, msg,
				G_DBUS_SEND_MESSAGE_FLAGS_NONE, NULL, &err)) {
		_E("Unable to send dbus message for releasing lock as  %s",
			err->message);
		ret = -1;
	}

	_D("Display lock released");
	g_object_unref(msg);
	g_dbus_connection_flush_sync(conn, NULL, NULL);
	g_clear_error(&err);
	return ret;
}

int _signal_send_system_service(int pid)
{
	int r;
	GVariant *param;

	param = g_variant_new("(i)", pid);
	if (!param) {
		_E("Out of memory");
		return -1;
	}

	r = __send_signal(RESOURCED_PROC_OBJECT,
			RESOURCED_PROC_INTERFACE,
			RESOURCED_SYSTEM_SERVICE_SIGNAL,
			param);
	if (r < 0) {
		_E("Failed to send system service signal - pid(%d)", pid);
		return -1;
	}

	_D("Send system service signal: pid(%d)", pid);

	return 0;
}

int _signal_subscribe_poweroff_state(void (*callback)(int state, void *data),
		void *user_data)
{
	if (callback == NULL)
		return -1;

	poweroff_state_sid = __subscribe_system_bus(SYSTEM_PATH_POWEROFF,
			SYSTEM_INTERFACE_POWEROFF,
			SYSTEM_POWEROFF_STATE_SIGNAL);
	if (poweroff_state_sid == 0) {
		_E("Failed to subscribe poweroff state signal");
		return -1;
	}

	poweroff_state_callback = callback;
	poweroff_state_data = user_data;
	_D("[SIGNAL] subscribe poweroff state");

	return 0;
}

int _signal_check_system_boot_finished(bool *finished)
{
	GError *err = NULL;
	GDBusMessage *msg;
	GDBusMessage *reply = NULL;
	GDBusConnection *conn;
	GVariant *body;
	GVariant *param;
	GVariant *v = NULL;
	gsize length = 0;
	const gchar *state;
	int ret = 0;

	if (!finished) {
		_E("Invalid parameter");
		return -1;
	}

	conn = __get_system_conn();
	if (!conn)
		return -1;

	msg = g_dbus_message_new_method_call(SD_BUS_NAME,
			SD_OBJECT_PATH,
			SD_PROPERTIES_INTERFACE,
			SD_GET_METHOD);
	if (!msg) {
		_E("g_dbus_message_new_method_call() is failed");
		return -1;
	}

	param = g_variant_new("(ss)",
			SD_MANAGER_INTERFACE,
			SD_SYSTEM_STATE_METHOD);
	g_dbus_message_set_body(msg, param);

	reply = g_dbus_connection_send_message_with_reply_sync(conn,
			msg,
			G_DBUS_SIGNAL_FLAGS_NONE,
			-1,
			NULL,
			NULL,
			&err);
	if (!reply) {
		_E("Failed to get reply. error(%s)", err->message);
		ret = -1;
		goto end;
	}

	body = g_dbus_message_get_body(reply);
	g_variant_get(body, "(v)", &v);
	if (!v) {
		_E("Failed to get variant");
		ret = -1;
		goto end;
	}

	state = g_variant_get_string(v, &length);
	if (g_strcmp0(state, "running") == 0 ||
			g_strcmp0(state, "degraded") == 0)
		*finished = true;
	else
		*finished = false;

	_W("state: %s, length: %d", state, (int)length);

end:
	if (v)
		g_variant_unref(v);
	if (reply)
		g_object_unref(reply);
	g_object_unref(msg);
	g_clear_error(&err);

	return ret;
}

static void __wm_proc_reply_cb(GObject *source_object,
		GAsyncResult *res, gpointer user_data)
{
	GDBusConnection *conn = (GDBusConnection *)source_object;
	proc_status_t *ps = (proc_status_t *)user_data;
	GDBusMessage *reply;
	GError *err = NULL;
	GVariant *var;
	int status = -1;
	int focus = -1;

	reply = g_dbus_connection_send_message_with_reply_finish(conn,
			res, &err);
	if (!reply) {
		_E("No reply. err(%s)", err ? err->message : "Unknown");
		goto end;
	}

	var = g_dbus_message_get_body(reply);
	if (!var) {
		_E("g_dbus_message_get_body() is failed");
		goto end;
	}

	g_variant_get(var, "(ii)", &status, &focus);
	if (status == -1 || focus == -1) {
		_E("Failed to get proc status info");
		goto end;
	}

	if (ps->callback)
		ps->callback(ps->pid, status, focus, ps->data);

	_D("pid(%d), status(%d), focused(%d)", ps->pid, status, focus);
end:
	if (err)
		g_error_free(err);
	if (reply)
		g_object_unref(reply);
	if (ps)
		free(ps);
}

int _signal_get_proc_status_async(int pid,
		void (*callback)(int, int, int, void *),
		void *data)
{
	GDBusConnection *conn;
	GDBusMessage *msg;
	proc_status_t *ps;

	conn = __get_system_conn();
	if (!conn)
		return -1;

	msg = g_dbus_message_new_method_call(WM_PROC_NAME,
			WM_PROC_PATH,
			WM_PROC_INTERFACE,
			WM_PROC_METHOD);
	if (!msg) {
		_E("g_dbus_message_new_method_call() is failed");
		return -1;
	}

	ps = calloc(1, sizeof(proc_status_t));
	if (!ps) {
		_E("Out of memory");
		g_object_unref(msg);
		return -1;
	}
	ps->pid = pid;
	ps->callback = callback;
	ps->data = data;

	g_dbus_message_set_body(msg, g_variant_new("(i)", pid));
	g_dbus_connection_send_message_with_reply(conn, msg,
			G_DBUS_SEND_MESSAGE_FLAGS_NONE,
			-1, NULL, NULL, __wm_proc_reply_cb,
			(gpointer)ps);
	g_object_unref(msg);

	return 0;
}

static gboolean __dispatch_ready_cb(gpointer user_data)
{
	ready_cb_info_t *info;
	GList *iter;

	iter = __ready_cbs;
	while (iter) {
		info = (ready_cb_info_t *)iter->data;
		iter = g_list_next(iter);
		if (info->callback(info->user_data) < 0)
			continue;

		__ready_cbs = g_list_remove(__ready_cbs, info);
		free(info);
	}

	if (!__ready_cbs) {
		_D("ready cb info is nullptr");
		return G_SOURCE_REMOVE;
	}

	return G_SOURCE_CONTINUE;
}

int _signal_add_ready_cb(signal_ready_cb callback, void *user_data)
{
	ready_cb_info_t *info;

	if (!callback) {
		_E("Invalid parameter");
		return -1;
	}

	info = malloc(sizeof(ready_cb_info_t));
	if (!info) {
		_E("Out of memory");
		return -1;
	}

	info->callback = callback;
	info->user_data = user_data;

	__ready_cbs = g_list_append(__ready_cbs, info);

	return 0;
}

static void __get_connection_cb(GObject *source_object,
		GAsyncResult *res, gpointer user_data)
{
	GError *error = NULL;

	if (system_conn) {
		_D("Already exists");
		return;
	}

	system_conn = g_bus_get_finish(res, &error);
	if (!system_conn) {
		_E("g_bus_get_finish() is failed. error(%s)", error->message);
		g_error_free(error);
	}

	if (__dispatch_ready_cb(NULL)) {
		g_timeout_add_seconds(RETRY_INTERVAL, __dispatch_ready_cb,
				NULL);
	}
}

int _signal_init(void)
{
	g_bus_get(G_BUS_TYPE_SYSTEM, NULL, __get_connection_cb, NULL);

	return 0;
}
